# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# Generated by Django 3.2.19 on 2025-08-19 13:52

import datetime

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0017_sosse_1_14_0"),
    ]

    operations = [
        migrations.CreateModel(
            name="CrawlPolicyBackup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "url_regex",
                    models.TextField(
                        help_text="URL regular expressions for this policy. (one by line, lines starting with # are ignored)"
                    ),
                ),
                ("url_regex_pg", models.TextField()),
                ("enabled", models.BooleanField(default=True)),
                (
                    "recursion",
                    models.CharField(
                        choices=[
                            ("always", "Crawl all pages"),
                            ("depth", "Depending on depth"),
                            ("never", "Never crawl"),
                        ],
                        default="always",
                        max_length=6,
                    ),
                ),
                ("mimetype_regex", models.TextField(default=".*")),
                (
                    "recursion_depth",
                    models.PositiveIntegerField(
                        default=0,
                        help_text="Level of external links (links that don't match the regex) to recurse into",
                    ),
                ),
                (
                    "keep_params",
                    models.BooleanField(
                        default=True,
                        help_text='When disabled, URL parameters (parameters after "?") are removed from URLs, this can be useful if some parameters are random, change sorting or filtering, ...',
                        verbose_name="Index URL parameters",
                    ),
                ),
                ("hide_documents", models.BooleanField(default=False, help_text="Hide documents from search results")),
                (
                    "default_browse_mode",
                    models.CharField(
                        choices=[
                            ("detect", "Detect"),
                            ("selenium", "Chromium"),
                            ("firefox", "Firefox"),
                            ("requests", "Python Requests"),
                        ],
                        default="detect",
                        help_text="Python Request is faster, but can't execute Javascript and may break pages",
                        max_length=8,
                    ),
                ),
                (
                    "snapshot_html",
                    models.BooleanField(
                        default=True,
                        help_text="Archive binary files, HTML content and download related assets",
                        verbose_name="Archive content 🔖",
                    ),
                ),
                (
                    "snapshot_exclude_url_re",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Regex of URL to skip related assets downloading",
                        verbose_name="Assets exclude URL regex",
                    ),
                ),
                (
                    "snapshot_exclude_mime_re",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Regex of mimetypes to skip related assets saving",
                        verbose_name="Assets exclude mime regex",
                    ),
                ),
                (
                    "snapshot_exclude_element_re",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Regex of HTML elements to skip related assets downloading",
                        verbose_name="Assets exclude HTML regex",
                    ),
                ),
                (
                    "thumbnail_mode",
                    models.CharField(
                        choices=[
                            ("preview", "Page preview from metadata"),
                            ("prevscreen", "Preview from meta, screenshot as fallback"),
                            ("screenshot", "Take a screenshot"),
                            ("none", "No thumbnail"),
                        ],
                        default="preview",
                        help_text="Save thumbnails to display in search results",
                        max_length=10,
                    ),
                ),
                (
                    "take_screenshots",
                    models.BooleanField(
                        default=False, help_text="Store pages as screenshots", verbose_name="Take screenshots 📷"
                    ),
                ),
                (
                    "screenshot_format",
                    models.CharField(choices=[("png", "png"), ("jpg", "jpg")], default="jpg", max_length=3),
                ),
                (
                    "remove_nav_elements",
                    models.CharField(
                        choices=[
                            ("idx", "From index"),
                            ("scr", "From index and screenshots"),
                            ("yes", "From index, screens and HTML snaps"),
                            ("no", "No"),
                        ],
                        default="idx",
                        help_text="Remove navigation related elements",
                        max_length=4,
                    ),
                ),
                (
                    "script",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Javascript code to execute after the page is loaded. If an object is returned, its content will be used to overwrite the document's fields",
                    ),
                ),
                (
                    "store_extern_links",
                    models.BooleanField(default=False, help_text="Store links to non-indexed pages"),
                ),
                (
                    "recrawl_freq",
                    models.CharField(
                        choices=[("none", "Once"), ("constant", "Constant time"), ("adaptive", "Adaptive")],
                        default="adaptive",
                        help_text="Adaptive frequency will increase delay between two crawls when the page stays unchanged",
                        max_length=8,
                        verbose_name="Crawl frequency",
                    ),
                ),
                (
                    "recrawl_dt_min",
                    models.DurationField(
                        blank=True,
                        default=datetime.timedelta(days=1),
                        help_text="Min. time before recrawling a page",
                        null=True,
                    ),
                ),
                (
                    "recrawl_dt_max",
                    models.DurationField(
                        blank=True,
                        default=datetime.timedelta(days=365),
                        help_text="Max. time before recrawling a page",
                        null=True,
                    ),
                ),
                (
                    "hash_mode",
                    models.CharField(
                        choices=[("raw", "Raw content"), ("no_numbers", "Normalize numbers")],
                        default="no_numbers",
                        help_text="Content to check for modifications",
                        max_length=10,
                        verbose_name="Change detection",
                    ),
                ),
                (
                    "recrawl_condition",
                    models.CharField(
                        choices=[
                            ("change", "On change only"),
                            ("always", "Always"),
                            ("manual", "On change or manual trigger"),
                        ],
                        default="manual",
                        help_text="Specifies the conditions under which a page is reprocessed",
                        max_length=10,
                        verbose_name="Condition",
                    ),
                ),
                (
                    "auth_login_url_re",
                    models.TextField(
                        blank=True,
                        help_text="A redirection to an URL matching the regex will trigger authentication",
                        null=True,
                        verbose_name="Login URL regex",
                    ),
                ),
                (
                    "auth_form_selector",
                    models.TextField(
                        blank=True,
                        help_text="CSS selector pointing to the authentication &lt;form&gt; element",
                        null=True,
                        verbose_name="Form selector",
                    ),
                ),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=128), blank=True, default=list, size=None
                    ),
                ),
                (
                    "webhooks",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=512), blank=True, default=list, size=None
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AuthFieldBackup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("key", models.CharField(max_length=256, verbose_name="<input> name attribute")),
                ("value", models.CharField(max_length=256)),
                (
                    "crawl_policy",
                    models.ForeignKey(on_delete=models.deletion.CASCADE, to="se.crawlpolicybackup"),
                ),
            ],
        ),
    ]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import json

from django.core.management.base import BaseCommand
from django.db import models
from django.utils.timezone import now

from ...document import Document
from ...models import WorkerStats


class Command(BaseCommand):
    help = "Display the crawling queue status."

    def add_arguments(self, parser):
        parser.add_argument("--format", choices=["text", "json"], default="text", help="Output format (text or json)")

    def handle(self, *args, **options):
        current_time = now()

        processing_count = Document.objects.wo_content().filter(worker_no__isnull=False).count()
        queue_new_count = Document.objects.wo_content().filter(crawl_last__isnull=True, worker_no__isnull=True).count()
        queue_recurring_count = (
            Document.objects.wo_content()
            .filter(
                crawl_last__isnull=False, crawl_next__isnull=False, crawl_next__lte=current_time, worker_no__isnull=True
            )
            .count()
        )

        queue_pending_count = (
            Document.objects.wo_content()
            .filter(models.Q(crawl_last__isnull=True) | models.Q(crawl_next__lte=current_time), worker_no__isnull=True)
            .count()
        )

        workers = WorkerStats.live_state()
        worker_count = workers.count()

        processing_docs = (
            Document.objects.wo_content().filter(worker_no__isnull=False).select_related().order_by("worker_no")
        )

        if options["format"] == "json":
            workers_data = []
            for worker in workers:
                worker_data = {
                    "worker_no": worker.worker_no,
                    "pid": worker.pid if worker.pid != "-" else None,
                    "state": worker.state,
                    "doc_processed": worker.doc_processed,
                    "current_url": None,
                }

                # Chercher le document en cours pour ce worker
                current_doc = processing_docs.filter(worker_no=worker.worker_no).first()
                if current_doc:
                    worker_data["current_url"] = current_doc.url

                workers_data.append(worker_data)

            result = {
                "timestamp": current_time.isoformat(),
                "queue": {
                    "processing": processing_count,
                    "new": queue_new_count,
                    "recurring": queue_recurring_count,
                    "total_pending": queue_pending_count,
                },
                "workers": {"count": worker_count, "details": workers_data},
            }

            self.stdout.write(json.dumps(result, indent=2, ensure_ascii=False))

        else:
            self.stdout.write(self.style.SUCCESS("=== Crawling Queue Status ==="))
            self.stdout.write(f"Timestamp: {current_time}")
            self.stdout.write("")

            self.stdout.write(self.style.WARNING("Queue:"))
            self.stdout.write(f"  Documents being processed: {processing_count}")
            self.stdout.write(f"  New documents pending: {queue_new_count}")
            self.stdout.write(f"  Recurring documents: {queue_recurring_count}")
            self.stdout.write(f"  Total documents in queue: {queue_pending_count}")
            self.stdout.write("")

            self.stdout.write(self.style.WARNING(f"Workers ({worker_count} total):"))

            if workers:
                for worker in workers:
                    state_color = (
                        self.style.SUCCESS
                        if worker.state == "running"
                        else (self.style.WARNING if worker.state == "idle" else self.style.ERROR)
                    )

                    self.stdout.write(f"  Worker {worker.worker_no}:")
                    self.stdout.write(f"    PID: {worker.pid}")
                    self.stdout.write(f"    State: {state_color(worker.state)}")
                    self.stdout.write(f"    Documents processed: {worker.doc_processed}")

                    # Display current processing URL
                    current_doc = processing_docs.filter(worker_no=worker.worker_no).first()
                    if current_doc:
                        self.stdout.write(f"    Current URL: {current_doc.url}")
                    else:
                        self.stdout.write("    Current URL: None")
                    self.stdout.write("")
            else:
                self.stdout.write("  No active workers")



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from unittest import TestCase

from django.core.management import call_command
from django.db import connection
from django.db.migrations.executor import MigrationExecutor

# Constants for old CrawlPolicy recursion values
CRAWL_ALL = "always"
CRAWL_ON_DEPTH = "depth"
CRAWL_NEVER = "never"


def migrate_to_state(app_label, migration_name):
    """Apply migration physically and return apps state."""
    call_command("migrate", app_label, migration_name, verbosity=0)
    executor = MigrationExecutor(connection)
    return executor.loader.project_state([(app_label, migration_name)]).apps


class MigrationExecutorTest(TestCase):
    @staticmethod
    def _clear_tables():
        with connection.cursor() as cursor:
            # Check if tables exist before truncating
            cursor.execute("""
                SELECT table_name FROM information_schema.tables
                WHERE table_schema = 'public' AND table_name IN
                ('se_document', 'se_crawlpolicy_tags', 'se_crawlpolicy_webhooks', 'se_crawlpolicy', 'se_authfield', 'se_tag', 'se_webhook')
            """)
            existing_tables = [row[0] for row in cursor.fetchall()]

            if existing_tables:
                tables_list = ", ".join(existing_tables)
                cursor.execute(f"TRUNCATE TABLE {tables_list} CASCADE")

    def setUp(self):
        MigrationExecutorTest._clear_tables()

    @classmethod
    def tearDownClass(cls):
        """Reapply all migrations and clean data after all tests."""
        super().tearDownClass()
        call_command("migrate", "se", verbosity=0)
        MigrationExecutorTest._clear_tables()

    def _test_migration_scenario(
        self,
        default_recursion,
        specific_url,
        specific_recursion,
        expected_default_unlimited,
        expected_default_limited,
        expected_specific_unlimited,
        expected_specific_limited,
    ):
        """Helper method to test migration scenarios with different recursion
        types."""
        # Migrate to state before Collection exists
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")

        # Since Django applies all migrations before the tests, we first clear all Crawl Policy to set a clean state
        CrawlPolicy.objects.all().delete()
        default_policy = CrawlPolicy.objects.create(
            url_regex="(default)", url_regex_pg=".*", recursion=default_recursion
        )
        CrawlPolicy.objects.create(url_regex=specific_url, url_regex_pg=specific_url, recursion=specific_recursion)
        Document = apps.get_model("se", "Document")
        Document.objects.create(url="http://default.com/")
        Document.objects.create(url=f"{specific_url}doc")

        # Migrate to final state with Collection
        apps = migrate_to_state("se", "0022_make_collection_non_null")
        Collection = apps.get_model("se", "Collection")
        Document = apps.get_model("se", "Document")

        # Should have 2 collections: Default and the specific one
        self.assertEqual(
            Collection.objects.count(), 2, Collection.objects.values("name", "unlimited_regex", "limited_regex")
        )

        # Check default collection
        default_collection = Collection.objects.get(name="Default")
        self.assertEqual(default_collection.unlimited_regex, expected_default_unlimited)
        self.assertEqual(default_collection.unlimited_regex_pg, expected_default_unlimited)
        self.assertEqual(default_collection.limited_regex, expected_default_limited)
        self.assertEqual(default_collection.limited_regex_pg, expected_default_limited)

        # Check specific collection
        specific_collection = Collection.objects.get(name=specific_url)
        self.assertEqual(specific_collection.unlimited_regex, expected_specific_unlimited)
        self.assertEqual(specific_collection.unlimited_regex_pg, expected_specific_unlimited)
        self.assertEqual(specific_collection.limited_regex, expected_specific_limited)
        self.assertEqual(specific_collection.limited_regex_pg, expected_specific_limited)

        # Check documents are assigned to correct collections
        default_doc = Document.objects.get(url="http://default.com/")
        specific_doc = Document.objects.get(url=f"{specific_url}doc")
        self.assertEqual(default_doc.collection, default_collection)
        self.assertEqual(specific_doc.collection, specific_collection)

        # Migrate back to initial state
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        self.assertEqual(CrawlPolicy.objects.count(), 2, CrawlPolicy.objects.values())

        # Check default policy was restored
        default_policy = CrawlPolicy.objects.get(url_regex="(default)")
        self.assertEqual(default_policy.url_regex, "(default)")
        self.assertEqual(default_policy.url_regex_pg, ".*")
        self.assertEqual(default_policy.recursion, default_recursion)

        # Check specific policy was restored
        specific_policy = CrawlPolicy.objects.get(url_regex=specific_url)
        self.assertEqual(specific_policy.url_regex, specific_url)
        self.assertEqual(specific_policy.url_regex_pg, specific_url)
        self.assertEqual(specific_policy.recursion, specific_recursion)

    def test_010_simple_migration(self):
        # Migrate to state before Collection exists
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")

        # Since Django applies all migrations before the tests, we first clear all Crawl Policy to set a clean state
        CrawlPolicy.objects.all().delete()
        CrawlPolicy.objects.create(url_regex="(default)", url_regex_pg=".*", recursion=CRAWL_ALL)

        # Migrate to final state with Collection
        apps = migrate_to_state("se", "0022_make_collection_non_null")
        Collection = apps.get_model("se", "Collection")
        self.assertEqual(
            Collection.objects.count(), 1, Collection.objects.values("name", "unlimited_regex", "unlimited_regex_pg")
        )
        collection = Collection.objects.first()
        self.assertEqual(collection.unlimited_regex, ".*")
        self.assertEqual(collection.unlimited_regex_pg, ".*")
        self.assertEqual(collection.limited_regex, "")
        self.assertEqual(collection.limited_regex_pg, "")

        # Migrate back to initial state
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        self.assertEqual(CrawlPolicy.objects.count(), 1, CrawlPolicy.objects.values())
        crawl_policy = CrawlPolicy.objects.first()
        self.assertEqual(crawl_policy.url_regex, "(default)")
        self.assertEqual(crawl_policy.url_regex_pg, ".*")
        self.assertEqual(crawl_policy.recursion, CRAWL_ALL)

    def test_020_default_never_policy_all(self):
        self._test_migration_scenario(CRAWL_NEVER, "http://test.com/", CRAWL_ALL, "", "", "http://test.com/", "")

    def test_030_default_never_policy_depth(self):
        self._test_migration_scenario(CRAWL_NEVER, "http://depth.com/", CRAWL_ON_DEPTH, "", "", "", "http://depth.com/")

    def test_040_default_depth_policy_all(self):
        self._test_migration_scenario(CRAWL_ON_DEPTH, "http://test.com/", CRAWL_ALL, "", ".*", "http://test.com/", ".*")

    def test_050_default_depth_policy_depth(self):
        self._test_migration_scenario(
            CRAWL_ON_DEPTH, "http://depth.com/", CRAWL_ON_DEPTH, "", ".*", "", "http://depth.com/"
        )

    def test_060_default_depth_policy_never(self):
        self._test_migration_scenario(CRAWL_ON_DEPTH, "http://never.com/", CRAWL_NEVER, "", ".*", "", "")

    def test_070_default_all_policy_all(self):
        self._test_migration_scenario(CRAWL_ALL, "http://test.com/", CRAWL_ALL, ".*", "", "http://test.com/", "")

    def test_080_default_all_policy_depth(self):
        self._test_migration_scenario(CRAWL_ALL, "http://depth.com/", CRAWL_ON_DEPTH, ".*", "", "", "http://depth.com/")

    def test_090_default_all_policy_never(self):
        self._test_migration_scenario(CRAWL_ALL, "http://never.com/", CRAWL_NEVER, ".*", "", "", "")

    def test_100_tags_migration(self):
        # Migrate to state before Collection exists
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        Tag = apps.get_model("se", "Tag")
        Document = apps.get_model("se", "Document")

        # Create test data
        CrawlPolicy.objects.all().delete()
        Tag.objects.all().delete()
        Document.objects.all().delete()

        tag1 = Tag.objects.create(name="tag1", depth=0, path="0001")
        tag2 = Tag.objects.create(name="tag2", depth=0, path="0002")

        default_policy = CrawlPolicy.objects.create(url_regex="(default)", url_regex_pg=".*", recursion=CRAWL_ALL)
        specific_policy = CrawlPolicy.objects.create(
            url_regex="http://test.com/", url_regex_pg="http://test.com/", recursion=CRAWL_ALL
        )

        default_policy.tags.set([tag1])
        specific_policy.tags.set([tag1, tag2])

        # Create documents and associate tags
        doc1 = Document.objects.create(url="http://default.com/")
        doc2 = Document.objects.create(url="http://test.com/doc")

        doc1.tags.set([tag1])
        doc2.tags.set([tag1, tag2])

        # Migrate to final state with Collection
        apps = migrate_to_state("se", "0022_make_collection_non_null")
        Collection = apps.get_model("se", "Collection")
        Document = apps.get_model("se", "Document")

        # Check tags are preserved on collections
        default_collection = Collection.objects.get(name="Default")
        specific_collection = Collection.objects.get(name="http://test.com/")

        self.assertEqual(list(default_collection.tags.values_list("name", flat=True)), ["tag1"])
        self.assertEqual(set(specific_collection.tags.values_list("name", flat=True)), {"tag1", "tag2"})

        # Check tags are preserved on documents
        doc1 = Document.objects.get(url="http://default.com/")
        doc2 = Document.objects.get(url="http://test.com/doc")

        self.assertEqual(list(doc1.tags.values_list("name", flat=True)), ["tag1"])
        self.assertEqual(set(doc2.tags.values_list("name", flat=True)), {"tag1", "tag2"})

        # Migrate back to initial state
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        Document = apps.get_model("se", "Document")

        # Check tags are restored on policies
        default_policy = CrawlPolicy.objects.get(url_regex="(default)")
        specific_policy = CrawlPolicy.objects.get(url_regex="http://test.com/")

        self.assertEqual(list(default_policy.tags.values_list("name", flat=True)), ["tag1"])
        self.assertEqual(set(specific_policy.tags.values_list("name", flat=True)), {"tag1", "tag2"})

        # Check tags are still on documents
        doc1 = Document.objects.get(url="http://default.com/")
        doc2 = Document.objects.get(url="http://test.com/doc")

        self.assertEqual(list(doc1.tags.values_list("name", flat=True)), ["tag1"])
        self.assertEqual(set(doc2.tags.values_list("name", flat=True)), {"tag1", "tag2"})

    def test_110_webhooks_migration(self):
        # Migrate to state before Collection exists
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        Webhook = apps.get_model("se", "Webhook")

        # Create test data
        CrawlPolicy.objects.all().delete()
        Webhook.objects.all().delete()

        webhook1 = Webhook.objects.create(name="webhook1", url="http://hook1.com")
        webhook2 = Webhook.objects.create(name="webhook2", url="http://hook2.com")

        default_policy = CrawlPolicy.objects.create(url_regex="(default)", url_regex_pg=".*", recursion=CRAWL_ALL)
        specific_policy = CrawlPolicy.objects.create(
            url_regex="http://test.com/", url_regex_pg="http://test.com/", recursion=CRAWL_ALL
        )

        default_policy.webhooks.set([webhook1])
        specific_policy.webhooks.set([webhook1, webhook2])

        # Migrate to final state with Collection
        apps = migrate_to_state("se", "0022_make_collection_non_null")
        Collection = apps.get_model("se", "Collection")
        Webhook = apps.get_model("se", "Webhook")

        # Check webhooks are preserved
        default_collection = Collection.objects.get(name="Default")
        specific_collection = Collection.objects.get(name="http://test.com/")

        self.assertEqual(list(default_collection.webhooks.values_list("name", flat=True)), ["webhook1"])
        self.assertEqual(set(specific_collection.webhooks.values_list("name", flat=True)), {"webhook1", "webhook2"})

        # Migrate back to initial state
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        Webhook = apps.get_model("se", "Webhook")

        # Check webhooks are restored
        default_policy = CrawlPolicy.objects.get(url_regex="(default)")
        specific_policy = CrawlPolicy.objects.get(url_regex="http://test.com/")

        self.assertEqual(list(default_policy.webhooks.values_list("name", flat=True)), ["webhook1"])
        self.assertEqual(set(specific_policy.webhooks.values_list("name", flat=True)), {"webhook1", "webhook2"})

    def test_120_authfields_migration(self):
        # Migrate to state before Collection exists
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        AuthField = apps.get_model("se", "AuthField")

        # Create test data
        CrawlPolicy.objects.all().delete()
        AuthField.objects.all().delete()

        default_policy = CrawlPolicy.objects.create(url_regex="(default)", url_regex_pg=".*", recursion=CRAWL_ALL)
        specific_policy = CrawlPolicy.objects.create(
            url_regex="http://test.com/", url_regex_pg="http://test.com/", recursion=CRAWL_ALL
        )

        AuthField.objects.create(key="username", value="admin", crawl_policy=default_policy)
        AuthField.objects.create(key="password", value="secret", crawl_policy=specific_policy)
        AuthField.objects.create(key="token", value="abc123", crawl_policy=specific_policy)

        # Migrate to final state with Collection
        apps = migrate_to_state("se", "0022_make_collection_non_null")
        Collection = apps.get_model("se", "Collection")
        AuthField = apps.get_model("se", "AuthField")

        # Check auth fields are preserved
        default_collection = Collection.objects.get(name="Default")
        specific_collection = Collection.objects.get(name="http://test.com/")

        default_auths = list(default_collection.authfield_set.values("key", "value"))
        specific_auths = list(specific_collection.authfield_set.values("key", "value"))

        self.assertEqual(default_auths, [{"key": "username", "value": "admin"}])
        self.assertEqual(len(specific_auths), 2)
        self.assertIn({"key": "password", "value": "secret"}, specific_auths)
        self.assertIn({"key": "token", "value": "abc123"}, specific_auths)

        # Migrate back to initial state
        apps = migrate_to_state("se", "0016_sosse_1_13_0_pre")
        CrawlPolicy = apps.get_model("se", "CrawlPolicy")
        AuthField = apps.get_model("se", "AuthField")

        # Check auth fields are restored
        default_policy = CrawlPolicy.objects.get(url_regex="(default)")
        specific_policy = CrawlPolicy.objects.get(url_regex="http://test.com/")

        default_auths = list(default_policy.authfield_set.values("key", "value"))
        specific_auths = list(specific_policy.authfield_set.values("key", "value"))

        self.assertEqual(default_auths, [{"key": "username", "value": "admin"}])
        self.assertEqual(len(specific_auths), 2)
        self.assertIn({"key": "password", "value": "secret"}, specific_auths)
        self.assertIn({"key": "token", "value": "abc123"}, specific_auths)



# Copyright 2024-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from unittest.mock import patch

from django.test import TestCase

from .add_to_queue import AddToQueueForm, _add_unique_patterns
from .collection import Collection
from .document import Document
from .test_views_mixin import ViewsTestMixin


class AddUniquePatternsFunctionTest(TestCase):
    def test_add_to_empty_regex(self):
        result = _add_unique_patterns("", "pattern1\npattern2")
        self.assertEqual(result, "pattern1\npattern2")

    def test_add_to_existing_regex(self):
        existing = "existing1\nexisting2"
        new_patterns = "pattern1\npattern2"
        result = _add_unique_patterns(existing, new_patterns)
        self.assertEqual(result, "existing1\nexisting2\npattern1\npattern2")

    def test_deduplicate_patterns(self):
        existing = "pattern1\npattern2"
        new_patterns = "pattern1\npattern3"
        result = _add_unique_patterns(existing, new_patterns)
        self.assertEqual(result, "pattern1\npattern2\npattern3")

    def test_empty_new_patterns(self):
        existing = "pattern1\npattern2"
        result = _add_unique_patterns(existing, "")
        self.assertEqual(result, "pattern1\npattern2")

    def test_empty_new_patterns_none(self):
        existing = "pattern1\npattern2"
        result = _add_unique_patterns(existing, None)
        self.assertEqual(result, "pattern1\npattern2")

    def test_all_duplicate_patterns(self):
        existing = "pattern1\npattern2"
        new_patterns = "pattern1\npattern2"
        result = _add_unique_patterns(existing, new_patterns)
        self.assertEqual(result, "pattern1\npattern2")

    def test_whitespace_handling(self):
        existing = "pattern1\npattern2"
        new_patterns = " pattern3 \n pattern4 \n\n"
        result = _add_unique_patterns(existing, new_patterns)
        self.assertEqual(result, "pattern1\npattern2\npattern3\npattern4")


class AddToQueueFormTest(TestCase):
    def setUp(self):
        self.collection = Collection.objects.create(name="Test Collection")

    def test_form_choices_constants(self):
        self.assertEqual(AddToQueueForm.CRAWL_SCOPE_NO_CHANGE, "no_change")
        self.assertEqual(AddToQueueForm.CRAWL_SCOPE_UNLIMITED, "unlimited")
        self.assertEqual(AddToQueueForm.CRAWL_SCOPE_LIMITED, "limited")

    def test_form_valid_data(self):
        form_data = {
            "urls": "http://example.com\nhttp://test.com",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_NO_CHANGE,
            "show_on_homepage": True,
        }
        form = AddToQueueForm(data=form_data)
        self.assertTrue(form.is_valid())

    def test_form_invalid_collection(self):
        form_data = {
            "urls": "http://example.com",
            "collection": 999999,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_NO_CHANGE,
        }
        form = AddToQueueForm(data=form_data)
        self.assertFalse(form.is_valid())

    def test_crawl_scope_choices(self):
        form = AddToQueueForm()
        expected_choices = [
            ("no_change", "Keep collection settings unchanged"),
            ("unlimited", "Crawl entire websites (unlimited depth)"),
            ("limited", "Crawl websites with depth limit from collection settings"),
        ]
        self.assertEqual(form.fields["crawl_scope"].choices, expected_choices)

    def test_crawl_scope_default(self):
        form = AddToQueueForm()
        self.assertEqual(form.fields["crawl_scope"].initial, AddToQueueForm.CRAWL_SCOPE_NO_CHANGE)


class AddToQueueViewTest(ViewsTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.collection.unlimited_regex = ""
        self.collection.limited_regex = ""
        self.collection.recursion_depth = 2
        self.collection.save()

    @patch.object(Document, "manual_queue")
    def test_form_valid_no_change(self, mock_queue):
        form_data = {
            "urls": "http://example.com",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_NO_CHANGE,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 302)

        mock_queue.assert_called_once_with("http://example.com/", self.collection, False)

        self.collection.refresh_from_db()
        self.assertEqual(self.collection.unlimited_regex, "")
        self.assertEqual(self.collection.limited_regex, "")

    @patch.object(Document, "manual_queue")
    def test_form_valid_unlimited_scope(self, mock_queue):
        form_data = {
            "urls": "http://example.com\nhttp://test.com",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_UNLIMITED,
            "show_on_homepage": True,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 302)

        self.assertEqual(mock_queue.call_count, 2)
        mock_queue.assert_any_call("http://example.com/", self.collection, True)
        mock_queue.assert_any_call("http://test.com/", self.collection, True)

        self.collection.refresh_from_db()
        patterns = self.collection.unlimited_regex.split("\n")
        self.assertIn("^https?://example\\.com/.*", patterns)
        self.assertIn("^https?://test\\.com/.*", patterns)
        self.assertEqual(len(patterns), 2)
        self.assertEqual(self.collection.limited_regex, "")

    @patch.object(Document, "manual_queue")
    def test_form_valid_limited_scope(self, mock_queue):
        form_data = {
            "urls": "http://example.com/path",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_LIMITED,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 302)

        mock_queue.assert_called_once_with("http://example.com/path", self.collection, False)

        self.collection.refresh_from_db()
        expected_pattern = "^https?://example\\.com/.*"
        self.assertEqual(self.collection.limited_regex, expected_pattern)
        self.assertEqual(self.collection.unlimited_regex, "")

    @patch.object(Document, "manual_queue")
    def test_form_valid_existing_patterns(self, mock_queue):
        self.collection.unlimited_regex = "^https?://existing\\.com/.*"
        self.collection.save()

        form_data = {
            "urls": "http://example.com",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_UNLIMITED,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 302)

        self.collection.refresh_from_db()
        expected_patterns = "^https?://existing\\.com/.*\n^https?://example\\.com/.*"
        self.assertEqual(self.collection.unlimited_regex, expected_patterns)

    @patch.object(Document, "manual_queue")
    def test_form_valid_duplicate_hostnames(self, mock_queue):
        form_data = {
            "urls": "http://example.com/page1\nhttp://example.com/page2",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_UNLIMITED,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 302)

        self.collection.refresh_from_db()
        expected_pattern = "^https?://example\\.com/.*"
        self.assertEqual(self.collection.unlimited_regex, expected_pattern)

    @patch.object(Document, "manual_queue")
    def test_form_invalid_url(self, mock_queue):
        form_data = {
            "urls": "invalid-url",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_UNLIMITED,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Invalid URL")

        mock_queue.assert_not_called()

        self.collection.refresh_from_db()
        self.assertEqual(self.collection.unlimited_regex, "")

    @patch.object(Document, "manual_queue")
    def test_form_invalid_relative_url(self, mock_queue):
        form_data = {
            "urls": "/relative/path",
            "collection": self.collection.id,
            "crawl_scope": AddToQueueForm.CRAWL_SCOPE_UNLIMITED,
            "show_on_homepage": False,
        }

        response = self.admin_client.post("/admin/se/document/queue/", form_data)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Invalid URL")

        mock_queue.assert_not_called()

        self.collection.refresh_from_db()
        self.assertEqual(self.collection.unlimited_regex, "")
        self.assertEqual(self.collection.limited_regex, "")

    def test_get_form(self):
        response = self.admin_client.get("/admin/se/document/queue/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'name="crawl_scope"')
        self.assertContains(response, "Keep collection settings unchanged")
        self.assertContains(response, "Crawl entire websites")
        self.assertContains(response, "Crawl websites with depth limit")



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import logging
import re
from datetime import timedelta

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import DataError, connection, models, transaction

from .browser import AuthElemFailed
from .browser_chromium import BrowserChromium
from .browser_firefox import BrowserFirefox
from .browser_request import BrowserRequest
from .document import Document
from .domain import Domain
from .tag import Tag
from .utils import build_multiline_re
from .webhook import Webhook

crawl_logger = logging.getLogger("crawler")
BROWSER_MAP = {
    Domain.BROWSE_CHROMIUM: BrowserChromium,
    Domain.BROWSE_FIREFOX: BrowserFirefox,
    Domain.BROWSE_REQUESTS: BrowserRequest,
}


@transaction.atomic
def validate_url_regexp(val):
    cursor = connection.cursor()
    for line_no, line in enumerate(val.splitlines()):
        line = line.strip()
        if line.startswith("#") or not line:
            continue

        try:
            # Try the regexp on Psql
            cursor.execute("SELECT 1 FROM se_document WHERE url ~ %s", params=[val])
        except DataError as e:
            if len(val.splitlines()) == 1:
                error = e.__cause__
            else:
                error = f"Regex on line {line_no + 1} failed: {e.__cause__}"
            raise ValidationError(error)


class Collection(models.Model):
    RECRAWL_FREQ_NONE = "none"
    RECRAWL_FREQ_CONSTANT = "constant"
    RECRAWL_FREQ_ADAPTIVE = "adaptive"
    RECRAWL_FREQ = [
        (RECRAWL_FREQ_NONE, "Once"),
        (RECRAWL_FREQ_CONSTANT, "Constant time"),
        (RECRAWL_FREQ_ADAPTIVE, "Adaptive"),
    ]

    HASH_RAW = "raw"
    HASH_NO_NUMBERS = "no_numbers"
    HASH_MODE = [
        (HASH_RAW, "Raw content"),
        (HASH_NO_NUMBERS, "Normalize numbers"),
    ]

    RECRAWL_COND_ON_CHANGE = "change"
    RECRAWL_COND_ALWAYS = "always"
    RECRAWL_COND_MANUAL = "manual"
    RECRAWL_CONDITION = [
        (RECRAWL_COND_ON_CHANGE, "On change only"),
        (RECRAWL_COND_ALWAYS, "Always"),
        (RECRAWL_COND_MANUAL, "On change or manual trigger"),
    ]

    REMOVE_NAV_FROM_INDEX = "idx"
    REMOVE_NAV_FROM_SCREENSHOT = "scr"
    REMOVE_NAV_FROM_ALL = "yes"
    REMOVE_NAV_NO = "no"
    REMOVE_NAV = [
        (REMOVE_NAV_FROM_INDEX, "From index"),
        (REMOVE_NAV_FROM_SCREENSHOT, "From index and screenshots"),
        (REMOVE_NAV_FROM_ALL, "From index, screens and HTML snaps"),
        (REMOVE_NAV_NO, "No"),
    ]

    THUMBNAIL_MODE_PREVIEW = "preview"
    THUMBNAIL_MODE_PREV_OR_SCREEN = "prevscreen"
    THUMBNAIL_MODE_SCREENSHOT = "screenshot"
    THUMBNAIL_MODE_NONE = "none"
    THUMBNAIL_MODE = (
        (THUMBNAIL_MODE_PREVIEW, "Page preview from metadata"),
        (THUMBNAIL_MODE_PREV_OR_SCREEN, "Preview from meta, screenshot as fallback"),
        (THUMBNAIL_MODE_SCREENSHOT, "Take a screenshot"),
        (THUMBNAIL_MODE_NONE, "No thumbnail"),
    )

    name = models.CharField(max_length=256, unique=True)
    unlimited_regex = models.TextField(
        blank=True,
        default="",
        validators=[validate_url_regexp],
        verbose_name="Unlimited depth URL regex",
        help_text="URL regular expressions. Matching URLs will have unlimited crawling recursion depth (one per line; lines starting with # are ignored)",
    )
    unlimited_regex_pg = models.TextField(default="")
    limited_regex = models.TextField(
        blank=True,
        default="",
        validators=[validate_url_regexp],
        verbose_name="Limited depth URL regex",
        help_text="URL regular expressions. Matching URLs will have limited crawling recursion depth (one per line; lines starting with # are ignored)",
    )
    limited_regex_pg = models.TextField(default="")
    combined_regex_pg = models.TextField(default="")
    excluded_regex = models.TextField(
        blank=True,
        default="",
        validators=[validate_url_regexp],
        verbose_name="Excluded URL regex",
        help_text="URL regular expressions to exclude from this collection. (one by line, lines starting with # are ignored)",
    )
    excluded_regex_pg = models.TextField(default="")
    mimetype_regex = models.TextField(default=".*")
    recursion_depth = models.PositiveIntegerField(
        default=1,
        verbose_name="Limited recursion depth",
        help_text="Level of external links (links that don't match the regex) to recurse into",
    )
    keep_params = models.BooleanField(
        default=True,
        verbose_name="Index URL parameters",
        help_text='When disabled, URL parameters (parameters after "?") are removed from URLs, this can be useful if some parameters are random, change sorting or filtering, ...',
    )
    hide_documents = models.BooleanField(default=False, help_text="Hide documents from search results")

    default_browse_mode = models.CharField(
        max_length=8,
        choices=Domain.BROWSE_MODE,
        default=Domain.BROWSE_DETECT,
        help_text="Python Request is faster, but can't execute Javascript and may break pages",
    )

    snapshot_html = models.BooleanField(
        default=True,
        help_text="Archive binary files, HTML content and download related assets",
        verbose_name="Archive content 🔖",
    )
    snapshot_exclude_url_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of URL to skip related assets downloading",
        verbose_name="Assets exclude URL regex",
    )
    snapshot_exclude_mime_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of mimetypes to skip related assets saving",
        verbose_name="Assets exclude mime regex",
    )
    snapshot_exclude_element_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of HTML elements to skip related assets downloading",
        verbose_name="Assets exclude HTML regex",
    )

    thumbnail_mode = models.CharField(
        default=THUMBNAIL_MODE_PREVIEW,
        help_text="Save thumbnails to display in search results",
        choices=THUMBNAIL_MODE,
        max_length=10,
    )
    take_screenshots = models.BooleanField(
        default=False,
        help_text="Store pages as screenshots",
        verbose_name="Take screenshots 📷",
    )
    screenshot_format = models.CharField(
        max_length=3,
        choices=Document.SCREENSHOT_FORMAT,
        default=Document.SCREENSHOT_JPG,
    )

    remove_nav_elements = models.CharField(
        default=REMOVE_NAV_FROM_INDEX,
        help_text="Remove navigation related elements",
        choices=REMOVE_NAV,
        max_length=4,
    )
    script = models.TextField(
        default="",
        help_text="Javascript code to execute after the page is loaded. If an object is returned, its content will be "
        "used to overwrite the document's fields",
        blank=True,
    )
    store_extern_links = models.BooleanField(default=False, help_text="Store links to non-indexed pages")

    recrawl_freq = models.CharField(
        max_length=8,
        choices=RECRAWL_FREQ,
        default=RECRAWL_FREQ_ADAPTIVE,
        verbose_name="Crawl frequency",
        help_text="Adaptive frequency will increase delay between two crawls when the page stays unchanged",
    )
    recrawl_dt_min = models.DurationField(
        blank=True,
        null=True,
        help_text="Min. time before recrawling a page",
        default=timedelta(days=1),
    )
    recrawl_dt_max = models.DurationField(
        blank=True,
        null=True,
        help_text="Max. time before recrawling a page",
        default=timedelta(days=365),
    )
    hash_mode = models.CharField(
        max_length=10,
        choices=HASH_MODE,
        default=HASH_NO_NUMBERS,
        verbose_name="Change detection",
        help_text="Content to check for modifications",
    )
    recrawl_condition = models.CharField(
        max_length=10,
        choices=RECRAWL_CONDITION,
        default=RECRAWL_COND_MANUAL,
        verbose_name="Condition",
        help_text="Specifies the conditions under which a page is reprocessed",
    )

    auth_login_url_re = models.TextField(
        null=True,
        blank=True,
        verbose_name="Login URL regex",
        help_text="A redirection to an URL matching the regex will trigger authentication",
    )
    auth_form_selector = models.TextField(
        null=True,
        blank=True,
        verbose_name="Form selector",
        help_text="CSS selector pointing to the authentication &lt;form&gt; element",
    )
    tags = models.ManyToManyField(Tag, blank=True)
    webhooks = models.ManyToManyField(Webhook)

    # Cross-collection crawl settings
    queue_to_any_collection = models.BooleanField(
        default=False,
        verbose_name="Queue to any collection",
        help_text="When a URL doesn't match this Collection's regex patterns, check all Collections and queue it in the first matching one.",
    )
    queue_to_collections = models.ManyToManyField(
        "self",
        blank=True,
        symmetrical=False,
        verbose_name="Queue to specific collections",
        help_text="When a URL doesn't match this Collection's regex patterns, check only these Collections and queue it there.",
    )

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.unlimited_regex_pg = build_multiline_re(self.unlimited_regex) if self.unlimited_regex else ""
        self.limited_regex_pg = build_multiline_re(self.limited_regex) if self.limited_regex else ""
        self.excluded_regex_pg = build_multiline_re(self.excluded_regex) if self.excluded_regex else ""

        # Build combined regex for cross-collection matching
        if self.unlimited_regex_pg and self.limited_regex_pg:
            self.combined_regex_pg = f"{self.unlimited_regex_pg}|{self.limited_regex_pg}"
        elif self.unlimited_regex_pg:
            self.combined_regex_pg = self.unlimited_regex_pg
        elif self.limited_regex_pg:
            self.combined_regex_pg = self.limited_regex_pg
        else:
            self.combined_regex_pg = ""

        return super().save(*args, **kwargs)

    @staticmethod
    def create_default():
        try:
            return Collection.objects.get(name="Default")
        except Collection.DoesNotExist:
            existing = Collection.objects.first()
            if existing:
                return existing
            return Collection.objects.create(name="Default")

    @staticmethod
    def get_from_url(url, collections_to_check=None):
        crawl_logger.debug("collections", Collection.objects.values_list("name", "combined_regex_pg"))
        collections = (
            Collection.objects.annotate(
                match_len=models.functions.Length(
                    models.Func(
                        models.Value(url),
                        models.F("combined_regex_pg"),
                        function="REGEXP_SUBSTR",
                        output_field=models.TextField(),
                    )
                )
            )
            .filter(match_len__gt=0)
            .order_by("-match_len")
        )

        # If specific collections are provided, filter to only those
        if collections_to_check is not None:
            collections_to_check_ids = {c.pk for c in collections_to_check}

        for collection in collections:
            # If we have a filter list, check if this collection is in it
            if collections_to_check is not None and collection.pk not in collections_to_check_ids:
                continue

            if not collection.combined_regex_pg:
                continue

            if Document._url_matches_regex(url, collection.excluded_regex_pg):
                continue
            return collection

        return None

    @staticmethod
    def _default_browser():
        if settings.SOSSE_DEFAULT_BROWSER == "chromium":
            return Domain.BROWSE_CHROMIUM
        return Domain.BROWSE_FIREFOX

    def url_get(self, url, domain=None):
        domain = domain or Domain.get_from_url(url)
        browser = self.get_browser(domain=domain, no_detection=False)
        page = browser.get(url, self)

        if page.redirect_count:
            # The request was redirected, check if we need auth
            try:
                crawl_logger.debug(f"may auth {page.url} / {self.auth_login_url_re}")
                if self.auth_login_url_re and self.auth_form_selector and re.search(self.auth_login_url_re, page.url):
                    crawl_logger.debug(f"doing auth for {url}")
                    new_page = page.browser.try_auth(page, url, self)

                    if new_page.url != url:
                        crawl_logger.debug(f"reopening {url} after auth")
                        page = browser.get(url, self)
                    else:
                        page = new_page
            except Exception as e:
                if isinstance(e, AuthElemFailed):
                    raise
                raise Exception("Authentication failed")

        if self.default_browse_mode == Domain.BROWSE_DETECT and domain.browse_mode == Domain.BROWSE_DETECT:
            crawl_logger.debug(f"browser detection on {url}")
            requests_page = BrowserRequest.get(url, self)
            browser_content = page.dom_walk(self, False, None)
            requests_content = requests_page.dom_walk(self, False, None)

            if browser_content["text"] != requests_content["text"]:
                new_mode = self._default_browser()
            else:
                new_mode = Domain.BROWSE_REQUESTS
                page = requests_page
            crawl_logger.debug(f"browser detected {new_mode} on {url}")
            domain.browse_mode = new_mode
            domain.save()
        return page

    def get_browser(self, url=None, domain=None, no_detection=True):
        if url is None and domain is None:
            raise Exception("Either url or domain must be provided")
        if url is not None and domain is not None:
            raise Exception("Either url or domain must be provided")

        if url:
            domain = Domain.get_from_url(url)

        browser_str = self.default_browse_mode
        if self.default_browse_mode == Domain.BROWSE_DETECT:
            if domain.browse_mode == Domain.BROWSE_DETECT:
                if no_detection:
                    raise Exception(f"browser mode is not yet known ({domain})")
                browser_str = self._default_browser()
            else:
                browser_str = domain.browse_mode

        return BROWSER_MAP[browser_str]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# Generated by Django 3.2.19 on 2025-08-19 13:52

from django.db import connection, migrations
from django.db.models.fields.reverse_related import (
    ManyToManyRel,
    ManyToOneRel,
    OneToOneRel,
)


def copy_model(source_model_name, target_model_name, apps):
    Tag = apps.get_model("se", "Tag")
    Webhook = apps.get_model("se", "Webhook")
    target_model = apps.get_model("se", target_model_name)
    source_model = apps.get_model("se", source_model_name)

    if source_model_name == "CrawlPolicy":
        source_auth_model = apps.get_model("se", "AuthField")
        target_auth_model = apps.get_model("se", "AuthFieldBackup")
    else:
        target_auth_model = apps.get_model("se", "AuthField")
        source_auth_model = apps.get_model("se", "AuthFieldBackup")

    target_auth_model.objects.all().delete()

    if source_model_name == "CrawlPolicy":
        target_model.objects.all().delete()
    else:
        # CrawlPolicy's definition is not available (in the code) we can't delete() it
        with connection.cursor() as cursor:
            cursor.execute("TRUNCATE TABLE se_crawlpolicy_tags, se_crawlpolicy_webhooks, se_crawlpolicy CASCADE;")

    fields = [
        field.name
        for field in apps.get_model("se", "CrawlPolicyBackup")._meta.get_fields()
        if field.name not in ("id", "tags", "webhooks")
        and not isinstance(field, (ManyToOneRel, ManyToManyRel, OneToOneRel))
    ]

    for source_instance in source_model.objects.all():
        target_instance = target_model()
        for field in fields:
            setattr(target_instance, field, getattr(source_instance, field))
        target_instance.save()

        for source_auth in source_auth_model.objects.filter(crawl_policy=source_instance):
            target_auth = target_auth_model()
            target_auth.key = source_auth.key
            target_auth.value = source_auth.value
            target_auth.crawl_policy_id = target_instance.id
            target_auth.save()

        if source_model_name == "CrawlPolicy":
            # Backup: save ManyToMany as ArrayField
            target_instance.tags = list(source_instance.tags.values_list("name", flat=True))
            target_instance.webhooks = list(source_instance.webhooks.values_list("name", flat=True))
            target_instance.save(update_fields=["tags", "webhooks"])
        else:
            # Restore: restore ArrayField to ManyToMany
            target_instance.tags.set(Tag.objects.filter(name__in=source_instance.tags))
            target_instance.webhooks.set(Webhook.objects.filter(name__in=source_instance.webhooks))


def backup_crawl_policy(apps, schema_editor):
    copy_model("CrawlPolicy", "CrawlPolicyBackup", apps)


def restore_crawl_policy(apps, schema_editor):
    copy_model("CrawlPolicyBackup", "CrawlPolicy", apps)


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0018_crawl_policy_backup"),
    ]

    operations = [
        migrations.RunPython(backup_crawl_policy, restore_crawl_policy),
    ]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0021_populate_document_collection"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collection",
            name="name",
            field=models.CharField(max_length=256, unique=True),
        ),
        migrations.AlterField(
            model_name="document",
            name="collection",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="se.collection"),
        ),
        migrations.RemoveField(
            model_name="collection",
            name="recursion",
        ),
        migrations.RemoveField(
            model_name="collection",
            name="enabled",
        ),
    ]



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from ...mime_plugin import MimePlugin
from ..builtin import UpdateBuiltinModel


class Command(UpdateBuiltinModel):
    help = "Updates MIME handler definitions."
    doc = "This updates MIME handlers in the database based on their definition in the filesystem."

    json_file = "sosse/mime_plugins.json"
    model_class = MimePlugin
    lookup_field = "name"
    model_name = "MIME handlers"
    fields_to_remove = {"dependencies"}



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# Generated by Django 3.2.19 on 2025-08-19 13:52

import django.db.models.deletion
from django.db import migrations, models

import se.collection
import se.url
import se.utils


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0019_crawl_policy_backup_run"),
    ]

    operations = [
        # Create backup of CrawlPolicy table before rename
        migrations.RenameModel(
            old_name="CrawlPolicy",
            new_name="Collection",
        ),
        migrations.RenameField(
            model_name="authfield",
            old_name="crawl_policy",
            new_name="collection",
        ),
        migrations.AddField(
            model_name="collection",
            name="name",
            field=models.CharField(max_length=256, unique=True, blank=True, null=True),
        ),
        migrations.RenameField(
            model_name="collection",
            old_name="url_regex",
            new_name="unlimited_regex",
        ),
        migrations.RenameField(
            model_name="collection",
            old_name="url_regex_pg",
            new_name="unlimited_regex_pg",
        ),
        migrations.AlterField(
            model_name="collection",
            name="unlimited_regex",
            field=models.TextField(
                blank=True,
                default="",
                help_text="URL regular expressions. Matching URLs will have unlimited crawling recursion depth (one per line; lines starting with # are ignored)",
                validators=[se.collection.validate_url_regexp],
                verbose_name="Unlimited depth URL regex",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="unlimited_regex_pg",
            field=models.TextField(default=""),
        ),
        migrations.AddField(
            model_name="collection",
            name="excluded_regex",
            field=models.TextField(
                blank=True,
                default="",
                help_text="URL regular expressions to exclude from this collection. (one by line, lines starting with # are ignored)",
                validators=[se.collection.validate_url_regexp],
                verbose_name="Excluded URL regex",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="excluded_regex_pg",
            field=models.TextField(default=""),
        ),
        migrations.AddField(
            model_name="collection",
            name="limited_regex",
            field=models.TextField(
                blank=True,
                default="",
                help_text="URL regular expressions. Matching URLs will have limited crawling recursion depth (one per line; lines starting with # are ignored)",
                validators=[se.collection.validate_url_regexp],
                verbose_name="Limited depth URL regex",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="limited_regex_pg",
            field=models.TextField(default=""),
        ),
        migrations.AddField(
            model_name="collection",
            name="combined_regex_pg",
            field=models.TextField(default=""),
        ),
        migrations.AlterField(
            model_name="collection",
            name="recursion_depth",
            field=models.PositiveIntegerField(
                default=1,
                help_text="Level of external links (links that don't match the regex) to recurse into",
                verbose_name="Limited recursion depth",
            ),
        ),
        migrations.AddField(
            model_name="document",
            name="collection",
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to="se.collection"),
        ),
        migrations.AlterField(
            model_name="document",
            name="url",
            field=models.TextField(validators=[se.url.validate_url]),
        ),
        migrations.AlterUniqueTogether(
            name="document",
            unique_together={("url", "collection")},
        ),
        migrations.AddIndex(
            model_name="document",
            index=models.Index(fields=["collection"], name="se_document_collect_f51e37_idx"),
        ),
        migrations.AlterField(
            model_name="collection",
            name="recursion_depth",
            field=models.PositiveIntegerField(
                default=1,
                help_text="Level of external links (links that don't match the regex) to recurse into",
                verbose_name="Limited recursion depth",
            ),
        ),
        migrations.AddField(
            model_name="webhook",
            name="url_re",
            field=models.TextField(
                blank=True,
                default=".*",
                help_text="Run the webhook on pages with URL matching this regex (one per line; lines starting with # are ignored)",
                verbose_name="URL regex",
                validators=[se.utils.validate_multiline_re],
            ),
        ),
        migrations.AlterModelOptions(
            name="collection",
            options={"verbose_name": "Collection", "verbose_name_plural": "Collections"},
        ),
        migrations.AlterModelOptions(
            name="collection",
            options={"ordering": ["name"]},
        ),
    ]



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import json
from io import StringIO

from django.core.management import call_command
from django.test import TransactionTestCase
from django.utils.timezone import now

from .collection import Collection
from .document import Document
from .models import WorkerStats


class CommandsTest(TransactionTestCase):
    def setUp(self):
        self.collection = Collection.create_default()
        Document.objects.wo_content().create(url="http://test/", collection=self.collection)

    def test_delete_document_match(self):
        self.assertEqual(Document.objects.count(), 1)
        call_command("delete_documents", "http://test")
        self.assertEqual(Document.objects.count(), 0)

    def test_delete_document_no_match(self):
        self.assertEqual(Document.objects.count(), 1)
        call_command("delete_documents", "http://no_test")
        self.assertEqual(Document.objects.count(), 1)

    def test_delete_document_dry_run(self):
        self.assertEqual(Document.objects.count(), 1)
        with self.assertRaises(SystemExit):
            call_command("delete_documents", "--dry-run", "http://test")
        self.assertEqual(Document.objects.count(), 1)

    def test_delete_document_exclude(self):
        # Create additional documents for exclusion test
        Document.objects.wo_content().create(url="http://test/important.html", collection=self.collection)
        Document.objects.wo_content().create(url="http://test/regular.html", collection=self.collection)
        self.assertEqual(Document.objects.wo_content().count(), 3)

        # Delete all test documents except those containing "important"
        call_command("delete_documents", "http://test", "--exclude", "important")

        # Should only have the important document left
        self.assertEqual(Document.objects.wo_content().count(), 1)
        remaining_doc = Document.objects.wo_content().first()
        self.assertIn("important", remaining_doc.url)

    def test_delete_document_ignore_case(self):
        # Create documents with different cases
        Document.objects.wo_content().create(url="http://TEST/page.html", collection=self.collection)
        Document.objects.wo_content().create(url="http://other/page.html", collection=self.collection)
        self.assertEqual(Document.objects.wo_content().count(), 3)

        # Delete with ignore case - should match both "test" and "TEST"
        call_command("delete_documents", "test", "--ignore-case")

        # Should only have the "other" document left
        self.assertEqual(Document.objects.wo_content().count(), 1)
        remaining_doc = Document.objects.wo_content().first()
        self.assertIn("other", remaining_doc.url)


class QueueStatusCommandTest(TransactionTestCase):
    def setUp(self):
        self.collection = Collection.create_default()
        self.current_time = now()

    def test_queue_status_text_format(self):
        """Test queue_status command with text output format."""
        # Create test data
        Document.objects.wo_content().create(url="http://new.test/", collection=self.collection)
        Document.objects.wo_content().create(url="http://processing.test/", collection=self.collection, worker_no=0)
        WorkerStats.objects.create(worker_no=0, pid=12345, state="idle", doc_processed=5)

        out = StringIO()
        call_command("queue_status", stdout=out)
        output = out.getvalue()

        # Verify basic structure
        self.assertIn("=== Crawling Queue Status ===", output)
        self.assertIn("Queue:", output)
        self.assertIn("Workers", output)
        self.assertIn("Documents being processed: 1", output)
        self.assertIn("New documents pending: 1", output)
        self.assertIn("Worker 0:", output)

    def test_queue_status_json_format(self):
        """Test queue_status command with JSON output format."""
        # Create test data
        Document.objects.wo_content().create(url="http://new.test/", collection=self.collection)
        WorkerStats.objects.create(worker_no=0, pid=12345, state="idle", doc_processed=5)

        out = StringIO()
        call_command("queue_status", "--format", "json", stdout=out)
        output = out.getvalue()

        # Parse and verify JSON structure
        data = json.loads(output)

        self.assertIn("timestamp", data)
        self.assertIn("queue", data)
        self.assertIn("workers", data)

        # Verify queue data
        queue_data = data["queue"]
        self.assertEqual(queue_data["processing"], 0)
        self.assertEqual(queue_data["new"], 1)
        self.assertEqual(queue_data["recurring"], 0)
        self.assertEqual(queue_data["total_pending"], 1)

        # Verify workers data structure
        workers_data = data["workers"]
        self.assertEqual(workers_data["count"], 1)
        self.assertEqual(len(workers_data["details"]), 1)

        worker_detail = workers_data["details"][0]
        self.assertEqual(worker_detail["worker_no"], 0)
        self.assertIn("pid", worker_detail)
        self.assertIn("state", worker_detail)
        self.assertEqual(worker_detail["doc_processed"], 5)



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django.contrib.auth.models import User
from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

from .collection import Collection
from .document import Document
from .move_to_collection import MoveToCollectionForm


class MoveToCollectionFormTest(TestCase):
    def setUp(self):
        self.collection1 = Collection.objects.create(name="Source Collection", unlimited_regex="http://source.com/.*")
        self.collection2 = Collection.objects.create(
            name="Destination Collection", unlimited_regex="http://dest.com/.*"
        )

    def test_form_fields(self):
        """Test that form has required fields with correct choices."""
        form = MoveToCollectionForm()

        # Test collection field
        self.assertIn("collection", form.fields)
        self.assertTrue(form.fields["collection"].required)

        # Test conflict_resolution field
        self.assertIn("conflict_resolution", form.fields)
        self.assertTrue(form.fields["conflict_resolution"].required)
        self.assertEqual(form.fields["conflict_resolution"].initial, MoveToCollectionForm.CONFLICT_OVERWRITE)

        # Test conflict choices
        expected_choices = [
            (MoveToCollectionForm.CONFLICT_SKIP, "Skip conflicting documents"),
            (MoveToCollectionForm.CONFLICT_OVERWRITE, "Overwrite documents in destination collection"),
            (MoveToCollectionForm.CONFLICT_DELETE_SOURCE, "Delete source document"),
        ]
        self.assertEqual(form.fields["conflict_resolution"].choices, expected_choices)

    def test_form_validation(self):
        """Test form validation with different inputs."""
        # Valid data
        form_data = {
            "collection": self.collection2.id,
            "conflict_resolution": MoveToCollectionForm.CONFLICT_OVERWRITE,
        }
        form = MoveToCollectionForm(data=form_data)
        self.assertTrue(form.is_valid())

        # Invalid conflict resolution
        form_data["conflict_resolution"] = "invalid_choice"
        form = MoveToCollectionForm(data=form_data)
        self.assertFalse(form.is_valid())

        # Missing collection
        form_data = {
            "conflict_resolution": MoveToCollectionForm.CONFLICT_SKIP,
        }
        form = MoveToCollectionForm(data=form_data)
        self.assertFalse(form.is_valid())


class MoveToCollectionViewTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser("admin", "admin@example.com", "password")
        self.collection1 = Collection.objects.create(name="Source Collection", unlimited_regex="http://source.com/.*")
        self.collection2 = Collection.objects.create(
            name="Destination Collection", unlimited_regex="http://dest.com/.*"
        )

        # Create test documents
        self.doc1 = Document.objects.create(
            url="http://example.com/page1",
            title="Document 1",
            content="Content 1",
            collection=self.collection1,
        )
        self.doc2 = Document.objects.create(
            url="http://example.com/page2",
            title="Document 2",
            content="Content 2",
            collection=self.collection1,
        )
        # Document with same URL in destination collection (conflict)
        self.doc_conflict = Document.objects.create(
            url="http://example.com/page1",
            title="Conflicting Document",
            content="Conflicting Content",
            collection=self.collection2,
        )

    def test_skip_conflicts(self):
        """Test skipping conflicting documents."""
        self.client.login(username="admin", password="password")

        # Set up session with documents to move
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        # Submit form with SKIP option
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection2.id,
                "conflict_resolution": MoveToCollectionForm.CONFLICT_SKIP,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check that doc1 was not moved (conflict), doc2 was moved
        self.doc1.refresh_from_db()
        self.doc2.refresh_from_db()

        self.assertEqual(self.doc1.collection, self.collection1)  # Not moved
        self.assertEqual(self.doc2.collection, self.collection2)  # Moved

        # Check that conflicting document still exists
        self.assertTrue(Document.objects.wo_content().filter(id=self.doc_conflict.id).exists())

    def test_overwrite_conflicts(self):
        """Test overwriting conflicting documents."""
        self.client.login(username="admin", password="password")

        # Set up session with documents to move
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        # Submit form with OVERWRITE option
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection2.id,
                "conflict_resolution": MoveToCollectionForm.CONFLICT_OVERWRITE,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check that both documents were moved
        self.doc1.refresh_from_db()
        self.doc2.refresh_from_db()

        self.assertEqual(self.doc1.collection, self.collection2)
        self.assertEqual(self.doc2.collection, self.collection2)

        # Check that conflicting document was deleted
        self.assertFalse(Document.objects.wo_content().filter(id=self.doc_conflict.id).exists())

    def test_delete_source_conflicts(self):
        """Test deleting source documents on conflict."""
        self.client.login(username="admin", password="password")

        # Set up session with documents to move
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        # Submit form with DELETE_SOURCE option
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection2.id,
                "conflict_resolution": MoveToCollectionForm.CONFLICT_DELETE_SOURCE,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check that doc1 was deleted (conflict), doc2 was moved
        self.assertFalse(Document.objects.wo_content().filter(id=self.doc1.id).exists())  # Deleted

        self.doc2.refresh_from_db()
        self.assertEqual(self.doc2.collection, self.collection2)  # Moved

        # Check that conflicting document still exists
        self.assertTrue(Document.objects.wo_content().filter(id=self.doc_conflict.id).exists())

    def test_no_conflicts(self):
        """Test moving documents without conflicts."""
        self.client.login(username="admin", password="password")

        # Remove the conflicting document
        self.doc_conflict.delete()

        # Set up session with documents to move
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        # Submit form
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection2.id,
                "conflict_resolution": MoveToCollectionForm.CONFLICT_OVERWRITE,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check that both documents were moved
        self.doc1.refresh_from_db()
        self.doc2.refresh_from_db()

        self.assertEqual(self.doc1.collection, self.collection2)
        self.assertEqual(self.doc2.collection, self.collection2)

    def test_no_documents_selected(self):
        """Test behavior when no documents are selected."""
        self.client.login(username="admin", password="password")

        # Submit form without setting documents_to_move in session
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection2.id,
                "conflict_resolution": MoveToCollectionForm.CONFLICT_OVERWRITE,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check error message
        messages = list(get_messages(response.wsgi_request))
        self.assertTrue(any("No documents selected" in str(message) for message in messages))

    def test_get_view_without_documents(self):
        """Test GET request without documents in session."""
        self.client.login(username="admin", password="password")

        response = self.client.get(reverse("admin:move_to_collection"))

        # Should redirect with error message
        self.assertEqual(response.status_code, 302)

        # Check error message
        messages = list(get_messages(response.wsgi_request))
        self.assertTrue(any("No documents selected" in str(message) for message in messages))

    def test_same_collection_as_destination(self):
        """Test moving documents to the same collection they are already in."""
        self.client.login(username="admin", password="password")

        # Set up session with documents to move to their current collection
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        # Submit form with same collection as source (collection1)
        response = self.client.post(
            reverse("admin:move_to_collection"),
            {
                "collection": self.collection1.id,  # Same as current collection
                "conflict_resolution": MoveToCollectionForm.CONFLICT_OVERWRITE,
            },
        )

        # Check redirect
        self.assertEqual(response.status_code, 302)

        # Check that documents remained in their original collection
        self.doc1.refresh_from_db()
        self.doc2.refresh_from_db()

        self.assertEqual(self.doc1.collection, self.collection1)
        self.assertEqual(self.doc2.collection, self.collection1)

    def test_get_view_with_documents(self):
        """Test GET request with documents in session."""
        self.client.login(username="admin", password="password")

        # Set up session with documents to move
        session = self.client.session
        session["documents_to_move"] = [self.doc1.id, self.doc2.id]
        session.save()

        response = self.client.get(reverse("admin:move_to_collection"))

        # Should display the form
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Move 2 documents to collection")
        self.assertContains(response, self.doc1.title)
        self.assertContains(response, self.doc2.title)



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# Generated by Django 4.2.23 on 2025-09-29 12:04

from django.db import migrations


def save_mime_plugins(apps, schema_editor):
    """Call save() on each MimePlugin to write scripts to disk."""
    MimePlugin = apps.get_model("se", "MimePlugin")
    for plugin in MimePlugin.objects.all():
        plugin.save()


def reverse_save_mime_plugins(apps, schema_editor):
    """Reverse operation - nothing to do."""
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0022_make_collection_non_null"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="MimeHandler",
            new_name="MimePlugin",
        ),
        migrations.RenameField(
            model_name="document",
            old_name="mime_handlers_result",
            new_name="mime_plugins_result",
        ),
        migrations.RunPython(
            save_mime_plugins,
            reverse_save_mime_plugins,
        ),
    ]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django.core.management.base import BaseCommand

from ...document import Document


class Command(BaseCommand):
    help = "Convert all PNG screenshots to JPEG format."

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Show what would be done without actually converting",
        )

    def handle(self, *args, **options):
        dry_run = options["dry_run"]

        if dry_run:
            self.stdout.write("DRY RUN: Showing what would be converted...")
        else:
            self.stdout.write("Converting PNG screenshots to JPEG...")

        documents_to_convert = (
            Document.objects.wo_content().exclude(screenshot_format=Document.SCREENSHOT_JPG).exclude(screenshot_count=0)
        )

        total_count = documents_to_convert.count()
        self.stdout.write(f"Found {total_count} documents with PNG screenshots to convert.")

        if total_count == 0:
            self.stdout.write("No documents need conversion.")
            return

        converted_count = 0

        for doc in documents_to_convert:
            if dry_run:
                self.stdout.write(f"Would convert: {doc.url}")
            else:
                try:
                    self.stdout.write(f"Converting: {doc.url}")
                    doc.convert_to_jpg()
                    doc.screenshot_format = Document.SCREENSHOT_JPG
                    doc.save()
                    converted_count += 1
                except Exception as e:
                    self.stdout.write(self.style.ERROR(f"Error converting {doc.url}: {e}"))

        if dry_run:
            self.stdout.write(f"DRY RUN: Would convert {total_count} documents.")
        else:
            self.stdout.write(
                self.style.SUCCESS(f"Successfully converted {converted_count} out of {total_count} documents.")
            )



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import os

from django.test import TransactionTestCase, override_settings

from .collection import Collection
from .document import Document
from .mime_plugin import MimePlugin
from .page import Page

TEST_CONTENT = "dumy data"


class MimePluginTests(TransactionTestCase):
    def setUp(self):
        self.collection = Collection.create_default()
        self.doc = Document.objects.create(mimetype="application/pdf", collection=self.collection, content="", error="")
        self.page = Page("http://127.0.0.1", TEST_CONTENT.encode("utf-8"), None)

    def test_json_doc_handler_execution(self):
        handler = MimePlugin.objects.create(
            name="PDF JSON Extractor",
            script='echo \'{"title": "My PDF"}\'',
            mimetype_re="^application/pdf$",
            enabled=True,
        )

        MimePlugin.run_for_document(self.doc, self.page)

        self.assertEqual(self.doc.content, "")
        self.assertEqual(self.doc.error, "")
        self.assertEqual(self.doc.title, "My PDF")
        self.assertTrue(os.path.exists(handler.get_script_path()))

    def test_handler_with_error_output(self):
        MimePlugin.objects.create(
            name="Failing Handler",
            script="exit 1",
            mimetype_re="^application/pdf$",
            enabled=True,
        )

        MimePlugin.run_for_document(self.doc, self.page)

        self.assertIn("Failing Handler", self.doc.error)
        self.assertEqual(self.doc.content, "")

    @override_settings(TEST_MODE=False)
    def test_handler_processing_error_json_invalid(self):
        MimePlugin.objects.create(
            name="Invalid JSON Handler",
            script="echo 'not a json'",
            mimetype_re="^application/pdf$",
            enabled=True,
        )

        MimePlugin.run_for_document(self.doc, self.page)

        self.assertIn("output is not valid JSON", self.doc.error)
        self.assertEqual(self.doc.content, "")

    def test_non_matching_mimetype(self):
        MimePlugin.objects.create(
            name="Other Type Handler",
            script="echo 'should not run'",
            mimetype_re="^image/.*$",
            enabled=True,
        )

        MimePlugin.run_for_document(self.doc, self.page)

        self.assertEqual(self.doc.content, "")
        self.assertEqual(self.doc.error, "")

    def test_enabled_field_prevents_execution(self):
        MimePlugin.objects.create(
            name="Disabled Handler",
            script="echo 'Should not run'",
            mimetype_re="^application/pdf$",
            enabled=False,
        )

        MimePlugin.run_for_document(self.doc, self.page)

        self.assertEqual(self.doc.content, "")
        self.assertEqual(self.doc.error, "")

    def test_mimetype_re_supports_multiple_patterns(self):
        MimePlugin.objects.create(
            name="Multiple Mimetypes Handler",
            script='echo \'{"content": "multi content"}\'',
            mimetype_re="^text/plain$\n^application/pdf$",
            enabled=True,
        )
        MimePlugin.run_for_document(self.doc, self.page)
        self.assertEqual(self.doc.content, "multi content")
        self.assertEqual(self.doc.error, "")

    def test_json_output(self):
        MimePlugin.objects.create(
            name="JSON Output Handler",
            script='echo \'{"title": "Output Format JSON"}\'',
            mimetype_re="^application/pdf$",
            enabled=True,
        )
        MimePlugin.run_for_document(self.doc, self.page)
        self.assertEqual(self.doc.title, "Output Format JSON")
        self.assertEqual(self.doc.error, "")

    def test_handler_timeout(self):
        MimePlugin.objects.create(
            name="Timeout Handler",
            script="sleep 5; echo fail",
            mimetype_re="^application/pdf$",
            timeout=1,
            enabled=True,
        )
        MimePlugin.run_for_document(self.doc, self.page)

        self.assertIn("Timeout", self.doc.error)
        self.assertEqual(self.doc.content, "")

    def test_handler_command_not_found(self):
        MimePlugin.objects.create(
            name="Missing Command",
            script="non_existent_command_xyz",
            mimetype_re="^application/pdf$",
            enabled=True,
        )
        MimePlugin.run_for_document(self.doc, self.page)
        self.assertIn("not found", self.doc.error)
        self.assertEqual(self.doc.content, "")

    def test_json_handler_with_input_file(self):
        MimePlugin.objects.create(
            name="Modify JSON Input",
            script=r"""#!/bin/bash
FILE="$1"
TITLE=$(jq -r '.title' "$FILE")
CONTENT=$(jq -r '.content' "$FILE")
URL=$(jq -r '.url' "$FILE")
echo "{\"title\": \"Modified $TITLE\", \"content\": \"Modified $CONTENT\"}"
""",
            mimetype_re="^application/pdf$",
            enabled=True,
        )
        self.doc.title = "Original Title"
        self.doc.content = "Original Content"
        self.doc.save()

        MimePlugin.run_for_document(self.doc, self.page)
        self.assertEqual(self.doc.title, "Modified Original Title")
        self.assertEqual(self.doc.content, "Modified Original Content")

    def test_json_handler_url_readonly(self):
        MimePlugin.objects.create(
            name="Modify JSON Input",
            script=r"""#!/bin/bash
FILE="$1"
TITLE=$(jq -r '.title' "$FILE")
CONTENT=$(jq -r '.content' "$FILE")
URL=$(jq -r '.url' "$FILE")
echo "{\"title\": \"Modified $TITLE\", \"content\": \"Modified $CONTENT\", \"url\": \"${URL}/modified\"}"
""",
            mimetype_re="^application/pdf$",
            enabled=True,
        )
        self.doc.title = "Original Title"
        self.doc.content = "Original Content"
        self.doc.url = "https://original/"
        self.doc.save()

        MimePlugin.run_for_document(self.doc, self.page)
        self.assertEqual(self.doc.title, "Modified Original Title")
        self.assertEqual(self.doc.content, "Modified Original Content")
        self.assertEqual(self.doc.url, "https://original/")

    def test_json_handler_normalized_fields(self):
        MimePlugin.objects.create(
            name="Modify JSON Input",
            script=r"""#!/bin/bash
FILE="$1"
TITLE=$(jq -r '.title' "$FILE")
CONTENT=$(jq -r '.content' "$FILE")
echo "{\"title\": \"Modifièd $TITLE\", \"content\": \"Modifiéd $CONTENT\"}"
""",
            mimetype_re="^application/pdf$",
            enabled=True,
        )
        self.doc.title = "Original Title"
        self.doc.content = "Original Content"
        self.doc.save()

        MimePlugin.run_for_document(self.doc, self.page)
        self.assertEqual(self.doc.title, "Modifièd Original Title")
        self.assertEqual(self.doc.normalized_title, "Modified Original Title")
        self.assertEqual(self.doc.content, "Modifiéd Original Content")
        self.assertEqual(self.doc.normalized_content, "Modified Original Content")



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import json
import os

from django.conf import settings
from django.core.management.base import BaseCommand


class UpdateBuiltinModel(BaseCommand):
    json_file = None
    model_class = None
    lookup_field = None
    model_name = None
    fields_to_remove = set()

    def handle(self, *args, **options):
        count = 0

        json_file_path = os.path.join(settings.BASE_DIR, self.json_file)
        for item in json.load(open(json_file_path, encoding="utf-8")):
            fields = item.copy()
            for field in self.fields_to_remove:
                fields.pop(field, None)

            lookup_value = fields.pop(self.lookup_field)

            db_obj, created = self.model_class.objects.get_or_create(
                **{self.lookup_field: lookup_value}, defaults=fields
            )

            if not created:
                self.update_existing(db_obj, fields)

            count += int(created)

        self.stdout.write(f"{count} new {self.model_name} added")
        self.check_conflicts()

    def update_existing(self, db_obj, fields):
        if not db_obj.builtin:
            self.stderr.write(
                self.style.WARNING(
                    f"Skipping update of {self.model_name.rstrip('s')} '{getattr(db_obj, self.lookup_field)}' "
                    f"because it is user-defined"
                )
            )
            return
        # Remove 'enabled' field to preserve user's enable/disable choice
        fields.pop("enabled", None)
        # Use save() instead of update() to ensure overridden save method is called
        # (e.g., MimePlugin needs to write the JSON configuration file to disk)
        for field, value in fields.items():
            setattr(db_obj, field, value)
        db_obj.save()

    def check_conflicts(self):
        pass



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import os
import re
import subprocess  # nosec B404
import tempfile
from mimetypes import guess_extension

from django.conf import settings
from django.db import models
from PIL import Image

from .builtin import BuiltinModel
from .utils import build_multiline_re, validate_multiline_re

crawl_logger = logging.getLogger("crawler")


class MimePlugin(BuiltinModel):
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True)
    license = models.CharField(max_length=255, blank=True, help_text="License of the script, e.g. GPL-3.0")
    script = models.TextField(help_text="Shell script which receives a file path as first argument")
    mimetype_re = models.TextField(
        help_text="One regex per line to match MIME types (e.g. ^application/pdf$)",
        validators=[validate_multiline_re],
    )
    timeout = models.PositiveIntegerField(default=30, help_text="Timeout in seconds for the script execution")
    enabled = models.BooleanField(default=True)

    def get_script_path(self):
        """Returns the absolute path to the stored script file, based on PK."""
        scripts_dir = settings.SOSSE_SCRIPTS_DIR
        os.makedirs(scripts_dir, exist_ok=True)
        return os.path.join(scripts_dir, f"plugin_{self.pk}.sh")

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        script_path = self.get_script_path()
        try:
            with open(script_path, "w") as f:
                script = self.script.replace("\r\n", "\n")
                f.write(script)
            os.chmod(script_path, 0o700)
            crawl_logger.debug(f"Script for handler '{self.name}' saved to {script_path}")
        except Exception as e:
            crawl_logger.error(f"Error saving script for handler '{self.name}': {e}")
            raise

    def delete(self, *args, **kwargs):
        script_path = self.get_script_path()
        if os.path.exists(script_path):
            try:
                os.remove(script_path)
                crawl_logger.debug(f"Deleted script file: {script_path}")
            except Exception as e:
                crawl_logger.error(f"Failed to delete script file {script_path}: {e}")
        return super().delete(*args, **kwargs)

    def execute(self, input_file: str, temp_dir: str):
        script_path = self.get_script_path()
        args = ["/bin/bash", script_path, input_file]

        crawl_logger.debug(f"[{self.name}] Executing script: {args}")

        try:
            result = subprocess.run(
                args,
                check=False,
                capture_output=True,
                text=True,
                cwd=temp_dir,
                timeout=self.timeout,
            )  # nosec B603
            crawl_logger.debug(f"[{self.name}] Return code: {result.returncode}")
            crawl_logger.debug(f"[{self.name}] STDOUT:\n{result.stdout.strip()}")
            crawl_logger.debug(f"[{self.name}] STDERR:\n{result.stderr.strip()}")
            return result.returncode, result.stdout.strip(), result.stderr.strip()
        except subprocess.TimeoutExpired as e:
            stderr_msg = e.stderr.decode() if e.stderr else ""
            msg = f"{stderr_msg}\n[{self.name}] Timeout after {self.timeout}s".strip()
            crawl_logger.debug(msg)
            return -1, "", msg

    @classmethod
    def run_for_document(cls, doc, page):
        from .rest_api import DocumentSerializer

        crawl_logger.debug(f"Running mime handlers for {doc.url}")

        applicable_handlers = cls.objects.filter(enabled=True).order_by("name")
        for handler in applicable_handlers:
            if re.match(build_multiline_re(handler.mimetype_re.strip()), doc.mimetype):
                break
        else:
            crawl_logger.debug(f"No handler matched for {doc.mimetype}")
            return

        temp_content_path = None
        temp_json_path = None
        try:
            # Write the content from the page if necessary
            content_file = doc.get_content_file()
            if not content_file:
                extension = guess_extension(doc.mimetype)
                with tempfile.NamedTemporaryFile(mode="wb+", suffix=f".{extension}", delete=False) as temp_file:
                    temp_file.write(page.content)
                    temp_file.flush()
                    temp_content_path = temp_file.name
                    content_file = temp_content_path

            with tempfile.NamedTemporaryFile("w", delete=False, suffix=".json") as tf:
                json_doc = DocumentSerializer(doc).data
                json_doc["content_file"] = content_file
                json.dump(json_doc, tf)
                tf.flush()
                temp_json_path = tf.name

            for handler in applicable_handlers:
                if re.match(build_multiline_re(handler.mimetype_re.strip()), doc.mimetype):
                    crawl_logger.debug(f"[{handler.name}] Handler matched for {doc.mimetype}")

                    with tempfile.TemporaryDirectory() as temp_dir:
                        code, stdout, stderr = handler.execute(temp_json_path, temp_dir)
                        doc.mime_plugins_result += f"{handler.name}:\n{stderr}\n\n"

                        if code != 0:
                            doc.error = f"{doc.error or ''}\nExecution of {handler.name} failed with status {code}:\n{stderr}".strip()
                            crawl_logger.error(f"Mime handler {handler.name} on {doc.url} with code {code}: {stderr}")
                            continue

                        try:
                            try:
                                data = json.loads(stdout)
                            except ValueError as e:
                                doc.error = f"{doc.error or ''}\n{handler.name} output is not valid JSON: {e}\ncontent: {stdout[:1000]}"
                                crawl_logger.error(
                                    f"Mime handler {handler.name} output is not valid JSON on {doc.url}: {e}\ncontent: {stdout[:1000]}"
                                )
                                continue

                            if stdout:
                                preview = None
                                if "preview" in data:
                                    preview = data.pop("preview")
                                    preview_file = os.path.join(temp_dir, preview)
                                    if os.path.exists(preview_file):
                                        with Image.open(preview_file) as img:
                                            # Remove alpha channel from the png
                                            img = img.convert("RGB")
                                            img.thumbnail((160, 100))
                                            thumb_jpg = os.path.join(
                                                settings.SOSSE_THUMBNAILS_DIR, doc.image_name() + ".jpg"
                                            )
                                            dir_name = os.path.dirname(thumb_jpg)
                                            os.makedirs(dir_name, exist_ok=True)
                                            img.save(thumb_jpg, "jpeg")
                                    else:
                                        raise Exception(f"Preview file {preview} does not exist")
                                serializer = DocumentSerializer(doc, data=data, partial=True)
                                serializer.is_valid(raise_exception=True)
                                serializer.update(doc, serializer.validated_data)
                                if preview:
                                    doc.has_thumbnail = True
                        except Exception as e:
                            doc.error = f"{doc.error or ''}\n{handler.name} processing error: {e}".strip()
                            doc.mime_plugins_result += f"{handler.name}:\n{stderr}\n\n"
                            crawl_logger.error(f"Mime handler {handler.name} processing error on {doc.url}: {e}")
                            if getattr(settings, "TEST_MODE", False):
                                raise
                            continue
        finally:
            if temp_json_path and os.path.exists(temp_json_path):
                os.remove(temp_json_path)
            if temp_content_path and os.path.exists(temp_content_path):
                os.remove(temp_content_path)

    def __str__(self):
        return self.name



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django.test import TransactionTestCase

from .archive import ArchiveMixin
from .collection import Collection
from .document import Document
from .test_views_mixin import ViewsTestMixin


class TestArchiveMixin(ArchiveMixin):
    """Test class that inherits from ArchiveMixin for testing purposes."""

    view_name = "test_view"


class ArchiveMixinTest(ViewsTestMixin, TransactionTestCase):
    def setUp(self):
        super().setUp()
        self.test_url = "https://example.com/test/page"

        # Create a second collection for testing
        self.collection2 = Collection.objects.create(name="Test Collection 2", unlimited_regex="https://example.com/.*")

        # Create documents in different collections
        self.doc1 = Document.objects.create(
            url=self.test_url, title="Test Document 1", collection=self.collection, content="Test content 1"
        )

        self.doc2 = Document.objects.create(
            url=self.test_url, title="Test Document 2", collection=self.collection2, content="Test content 2"
        )

    def test_get_document_without_collection_in_url(self):
        """Test that _get_document returns the first document when no
        collection is specified in URL."""
        # Create a request without collection ID in the path
        request = self.factory.get(f"/www/{self.test_url}")
        request.META["REQUEST_URI"] = f"/www/{self.test_url}"

        mixin = TestArchiveMixin()
        mixin.request = request

        # Should return the first document (ordered by id)
        doc = mixin._get_document()
        self.assertIsNotNone(doc)
        self.assertEqual(doc.id, self.doc1.id)
        self.assertEqual(doc.title, "Test Document 1")
        self.assertIsNone(mixin.collection_id)
        self.assertIsNone(mixin.collection)

    def test_get_document_with_collection_in_url(self):
        """Test that _get_document returns the document from the specified
        collection when collection ID is in URL."""
        # Create a request with collection ID in the path
        request = self.factory.get(f"/www/{self.collection2.id}/{self.test_url}")
        request.META["REQUEST_URI"] = f"/www/{self.collection2.id}/{self.test_url}"

        mixin = TestArchiveMixin()
        mixin.request = request

        # Should return the document from collection2
        doc = mixin._get_document()
        self.assertIsNotNone(doc)
        self.assertEqual(doc.id, self.doc2.id)
        self.assertEqual(doc.title, "Test Document 2")
        self.assertEqual(doc.collection, self.collection2)
        self.assertEqual(mixin.collection_id, self.collection2.id)
        self.assertEqual(mixin.collection, self.collection2)



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django import forms
from django.contrib import messages
from django.db import transaction
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import FormView

from .collection import Collection
from .document import Document
from .views import AdminView


class MoveToCollectionForm(forms.Form):
    CONFLICT_SKIP = "skip"
    CONFLICT_OVERWRITE = "overwrite"
    CONFLICT_DELETE_SOURCE = "delete_source"

    CONFLICT_CHOICES = [
        (CONFLICT_SKIP, "Skip conflicting documents"),
        (CONFLICT_OVERWRITE, "Overwrite documents in destination collection"),
        (CONFLICT_DELETE_SOURCE, "Delete source document"),
    ]

    collection = forms.ModelChoiceField(
        queryset=Collection.objects.all(),
        required=True,
        label="Destination Collection",
        widget=forms.Select(attrs={"class": "form-control"}),
        empty_label=None,
    )

    conflict_resolution = forms.ChoiceField(
        choices=CONFLICT_CHOICES,
        initial=CONFLICT_OVERWRITE,
        required=True,
        label="When URL conflicts occur",
        widget=forms.Select(attrs={"class": "form-control"}),
    )


class MoveToCollectionView(AdminView, FormView):
    template_name = "admin/move_to_collection.html"
    title = "Move Documents to Collection"
    form_class = MoveToCollectionForm
    permission_required = "se.change_document"
    admin_site = None

    def __init__(self, *args, **kwargs):
        self.admin_site = kwargs.pop("admin_site")
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.admin_site.each_context(self.request))

        document_ids = self.request.session.get("documents_to_move", [])
        documents = Document.objects.wo_content().filter(id__in=document_ids)
        document_count = documents.count()

        context.update(
            {
                "documents": documents,
                "document_count": document_count,
                "return_url": reverse("admin:se_document_changelist"),
            }
        )
        return context

    def form_valid(self, form):
        document_ids = self.request.session.get("documents_to_move", [])
        if not document_ids:
            messages.error(self.request, "No documents selected.")
            return redirect(reverse("admin:se_document_changelist"))

        destination_collection = form.cleaned_data["collection"]
        conflict_resolution = form.cleaned_data["conflict_resolution"]
        documents = Document.objects.wo_content().filter(id__in=document_ids)
        document_count = documents.count()

        if document_count == 0:
            messages.error(self.request, "No documents found to move.")
        else:
            moved_count = 0
            skipped_count = 0
            overwritten_count = 0
            deleted_count = 0

            with transaction.atomic():
                for document in documents:
                    # Skip documents that are already in the destination collection
                    if document.collection == destination_collection:
                        continue

                    # Check if a document with the same URL already exists in destination collection
                    existing_doc = (
                        Document.objects.wo_content()
                        .filter(url=document.url, collection=destination_collection)
                        .first()
                    )

                    if existing_doc:
                        if conflict_resolution == MoveToCollectionForm.CONFLICT_SKIP:
                            skipped_count += 1
                            continue
                        elif conflict_resolution == MoveToCollectionForm.CONFLICT_OVERWRITE:
                            # Delete the existing document first
                            existing_doc.delete()
                            overwritten_count += 1
                        elif conflict_resolution == MoveToCollectionForm.CONFLICT_DELETE_SOURCE:
                            # Delete the source document instead of moving it
                            document.delete()
                            deleted_count += 1
                            continue

                    # Move the document
                    document.collection = destination_collection
                    document.save()
                    moved_count += 1

            # Build success message
            message_parts = []
            if moved_count > 0:
                message_parts.append(f"{moved_count} document{'s' if moved_count > 1 else ''} moved")
            if overwritten_count > 0:
                message_parts.append(f"{overwritten_count} document{'s' if overwritten_count > 1 else ''} overwritten")
            if deleted_count > 0:
                message_parts.append(f"{deleted_count} source document{'s' if deleted_count > 1 else ''} deleted")
            if skipped_count > 0:
                message_parts.append(
                    f"{skipped_count} document{'s' if skipped_count > 1 else ''} skipped due to conflicts"
                )

            if message_parts:
                message = (
                    f"Operation completed: {', '.join(message_parts)} to collection '{destination_collection.name}'."
                )
                messages.success(self.request, message)
            else:
                messages.warning(self.request, "No documents were processed.")

        del self.request.session["documents_to_move"]
        return redirect(reverse("admin:se_document_changelist"))

    def get(self, request, *args, **kwargs):
        document_ids = request.session.get("documents_to_move", [])
        if not document_ids:
            messages.error(request, "No documents selected.")
            return redirect(reverse("admin:se_document_changelist"))
        return super().get(request, *args, **kwargs)



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# Generated by Django 3.2.19 on 2025-06-20 06:55

from django.db import migrations, models
from django.db.models import functions

import se.utils


def truncate_mimetype_reverse(apps, schema_editor):
    """Truncate mimetype field values to 64 chars for reverse migration."""
    Document = apps.get_model("se", "Document")
    Document.objects.filter(mimetype__isnull=False).update(mimetype=functions.Left("mimetype", 64))


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0016_sosse_1_13_0_pre"),
    ]

    operations = [
        migrations.CreateModel(
            name="MimeHandler",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True)),
                (
                    "license",
                    models.CharField(blank=True, help_text="License of the script, e.g. GPL-3.0", max_length=255),
                ),
                ("script", models.TextField(help_text="Shell script which receives a file path as first argument")),
                (
                    "mimetype_re",
                    models.TextField(
                        help_text="One regex per line to match MIME types (e.g. ^application/pdf$)",
                        validators=[se.utils.validate_multiline_re],
                    ),
                ),
                (
                    "timeout",
                    models.PositiveIntegerField(default=30, help_text="Timeout in seconds for the script execution"),
                ),
                ("builtin", models.BooleanField(default=False, verbose_name="Built-in")),
                ("enabled", models.BooleanField(default=True)),
            ],
        ),
        migrations.RenameModel(
            old_name="DomainSetting",
            new_name="Domain",
        ),
        migrations.AddField(
            model_name="document",
            name="mime_handlers_result",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AddField(
            model_name="document",
            name="retries",
            field=models.PositiveIntegerField(default=0, verbose_name="Crawl retries"),
        ),
        migrations.AlterField(
            model_name="document",
            name="mimetype",
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.RunPython(
            migrations.RunPython.noop,
            truncate_mimetype_reverse,
        ),
    ]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

# This function is imported from urllib3 and monkey-patched to avoid https://github.com/urllib3/urllib3/issues/3081
# MIT License
# Copyright (c) 2008-2020 Andrey Petrov and contributors.

# Detect urllib3 version for backward compatibility
try:
    import urllib3

    URLLIB3_VERSION_MAJOR = int(urllib3.__version__.split(".")[0])
except (ImportError, AttributeError, ValueError, IndexError):
    URLLIB3_VERSION_MAJOR = 1

# Import the appropriate modules based on urllib3 version
if URLLIB3_VERSION_MAJOR < 2:
    # urllib3 1.x compatibility (old API)
    import six
    from urllib3.util.url import PERCENT_RE

    def _encode_invalid_chars(component, allowed_chars, encoding="utf-8"):
        """Percent-encodes a URI component without reapplying onto an already
        percent-encoded component."""
        if component is None:
            return component

        component = six.ensure_text(component)

        # Normalize existing percent-encoded bytes.
        # Try to see if the component we're encoding is already percent-encoded
        # so we can skip all '%' characters but still encode all others.
        component, percent_encodings = PERCENT_RE.subn(lambda match: match.group(0), component)

        uri_bytes = component.encode("utf-8", "surrogatepass")
        is_percent_encoded = percent_encodings == uri_bytes.count(b"%")
        encoded_component = bytearray()

        for i in range(0, len(uri_bytes)):
            # Will return a single character bytestring on both Python 2 & 3
            byte = uri_bytes[i : i + 1]
            byte_ord = ord(byte)
            if (is_percent_encoded and byte == b"%") or (byte_ord < 128 and byte.decode() in allowed_chars):
                encoded_component += byte
                continue
            encoded_component.extend(b"%" + (hex(byte_ord)[2:].encode().zfill(2).upper()))

        return encoded_component.decode(encoding)

else:
    # urllib3 2.x+ (new API)
    import typing

    from urllib3.util.url import _PERCENT_RE
    from urllib3.util.util import to_str

    def _encode_invalid_chars(component: str | None, allowed_chars: typing.Container[str]) -> str | None:
        """Percent-encodes a URI component without reapplying onto an already
        percent-encoded component."""
        if component is None:
            return component

        component = to_str(component)

        # Normalize existing percent-encoded bytes.
        # Try to see if the component we're encoding is already percent-encoded
        # so we can skip all '%' characters but still encode all others.
        component, percent_encodings = _PERCENT_RE.subn(lambda match: match.group(0), component)

        uri_bytes = component.encode("utf-8", "surrogatepass")
        is_percent_encoded = percent_encodings == uri_bytes.count(b"%")
        encoded_component = bytearray()

        for i in range(0, len(uri_bytes)):
            # Will return a single character bytestring
            byte = uri_bytes[i : i + 1]
            byte_ord = ord(byte)
            if (is_percent_encoded and byte == b"%") or (byte_ord < 128 and byte.decode() in allowed_chars):
                encoded_component += byte
                continue
            encoded_component.extend(b"%" + (hex(byte_ord)[2:].encode().zfill(2).upper()))

        return encoded_component.decode()



# Copyright 2024-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django.core.exceptions import ValidationError
from django.test import TransactionTestCase

from .collection import Collection


class CollectionTest(TransactionTestCase):
    def setUp(self):
        self.default_policy = Collection.create_default()
        self.policy = Collection.objects.create(name="Test Policy", unlimited_regex="http://127.0.0.1/")

    def test_010_default(self):
        default = Collection.create_default()
        self.assertEqual(self.default_policy, default)

    def test_020_unlimited_regex_parse(self):
        empty = Collection.objects.create(name="Empty Collection", unlimited_regex="# Empty re")
        self.assertEqual(empty.unlimited_regex_pg, "")

        one_re = Collection.objects.create(name="One Regex Collection", unlimited_regex="http://.*")
        self.assertEqual(one_re.unlimited_regex_pg, "http://.*")

        multi_re = Collection.objects.create(name="Multi Regex Collection", unlimited_regex="http://a.*\nhttp://b.*")
        self.assertEqual(multi_re.unlimited_regex_pg, "(http://a.*|http://b.*)")

        other_one_re = Collection.objects.create(
            name="Other Collection",
            limited_regex="http://.*",
            excluded_regex="http://.*",
        )
        self.assertEqual(other_one_re.limited_regex_pg, "http://.*")
        self.assertEqual(other_one_re.excluded_regex_pg, "http://.*")

    def test_030_match(self):
        policy = Collection.get_from_url("http://127.0.0.1/")
        self.assertEqual(policy, self.policy)

    def test_040_no_match(self):
        Collection.get_from_url("# Empty")
        policy = Collection.get_from_url("http://127.0.0.2/")
        self.assertIsNone(policy)

    def test_050_longest_match(self):
        sub_policy = Collection.objects.create(name="Sub Policy Collection", unlimited_regex="http://127.0.0.1/abc/")

        policy = Collection.get_from_url("http://127.0.0.1/abc/plop")
        self.assertEqual(policy, sub_policy)

        policy = Collection.get_from_url("http://127.0.0.1/blah")
        self.assertEqual(policy, self.policy)

        longer = Collection.objects.create(name="Longer Collection", unlimited_regex="http://127.0.0.1/.*")
        policy = Collection.get_from_url("http://127.0.0.1/abc/blah")
        self.assertEqual(policy, longer)

    def test_060_multi_match(self):
        multi = Collection.objects.create(
            name="Multi Match Collection", unlimited_regex="http://127.0.0.1/match1/\nhttp://127.0.0.1/match2/"
        )

        policy = Collection.get_from_url("http://127.0.0.1/match1/url")
        self.assertEqual(policy, multi)

        policy = Collection.get_from_url("http://127.0.0.1/match2/url")
        self.assertEqual(policy, multi)

    def test_070_invalid_regex(self):
        with self.assertRaises(ValidationError):
            policy = Collection(name="Invalid Collection", unlimited_regex="(")
            policy.full_clean()

    def test_080_combined_regex_pg(self):
        # Test with unlimited_regex only
        unlimited_only = Collection.objects.create(name="Unlimited Only", unlimited_regex="http://example.com/")
        self.assertEqual(unlimited_only.combined_regex_pg, "http://example.com/")

        # Test with limited_regex only
        limited_only = Collection.objects.create(name="Limited Only", limited_regex="http://test.com/")
        self.assertEqual(limited_only.combined_regex_pg, "http://test.com/")

        # Test with both unlimited and limited
        both = Collection.objects.create(
            name="Both", unlimited_regex="http://example.com/", limited_regex="http://test.com/"
        )
        self.assertEqual(both.combined_regex_pg, "http://example.com/|http://test.com/")

        # Test with empty regex
        empty = Collection.objects.create(name="Empty", unlimited_regex="", limited_regex="")
        self.assertEqual(empty.combined_regex_pg, "")

    def test_090_get_from_url_with_combined_regex(self):
        # Create collections with different regex patterns
        coll1 = Collection.objects.create(name="Collection 1", unlimited_regex="http://site1.com/")
        coll2 = Collection.objects.create(name="Collection 2", limited_regex="http://site2.com/")
        coll3 = Collection.objects.create(
            name="Collection 3", unlimited_regex="http://site3.com/", limited_regex="http://site3.com/api/"
        )

        # Test matching with unlimited_regex only
        policy = Collection.get_from_url("http://site1.com/page")
        self.assertEqual(policy, coll1)

        # Test matching with limited_regex only
        policy = Collection.get_from_url("http://site2.com/page")
        self.assertEqual(policy, coll2)

        # Test matching with combined regex (should match longer match)
        policy = Collection.get_from_url("http://site3.com/api/endpoint")
        self.assertEqual(policy, coll3)

        # Test no match
        policy = Collection.get_from_url("http://nomatch.com/")
        self.assertIsNone(policy)



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import json
from collections import namedtuple
from unittest import mock

from django.contrib.auth.models import User
from django.test import Client, TransactionTestCase
from django.utils import timezone

from .collection import Collection
from .document import Document
from .mime_plugin import MimePlugin
from .models import CrawlerStats
from .tag import Tag
from .webhook import Webhook

now = timezone.now()
now_str = now.isoformat().replace("+00:00", "Z")
SERIALIZED_DOC1 = {
    "content": "Content",
    "content_hash": None,
    "crawl_dt": None,
    "crawl_first": now_str,
    "crawl_last": now_str,
    "crawl_next": None,
    "crawl_recurse": 0,
    "error": "",
    "error_hash": "",
    "favicon": None,
    "has_html_snapshot": False,
    "has_thumbnail": False,
    "hidden": False,
    "lang_iso_639_1": "en",
    "manual_crawl": False,
    "metadata": {},
    "mime_plugins_result": "",
    "mimetype": "text/html",
    "modified_date": None,
    "normalized_content": "content",
    "normalized_title": "title",
    "normalized_url": "http test",
    "redirect_url": None,
    "retries": 0,
    "robotstxt_rejected": False,
    "screenshot_count": 0,
    "screenshot_format": "",
    "screenshot_size": "",
    "show_on_homepage": False,
    "tags": [],
    "tags_str": "",
    "title": "Title",
    "too_many_redirects": False,
    "url": "http://127.0.0.1/test",
    "vector": "'content':4C 'http':2A 'test':3A 'title':1A",
    "vector_lang": "simple",
    "webhooks_result": {},
    "worker_no": None,
}
SERIALIZED_DOC2 = SERIALIZED_DOC1 | {
    "url": "http://127.0.0.1/test2",
    "normalized_url": "http test2",
    "title": "Title2",
    "normalized_title": "title2",
    "content": "Other Content2",
    "normalized_content": "other content2",
    "mimetype": "image/png",
    "tags": ["Sub Tag"],
    "tags_str": "Sub Tag",
    "vector": "'content2':5C 'http':2A 'other':4C 'test2':3A 'title2':1A",
}
SERIALIZED_DOC3 = SERIALIZED_DOC1 | {
    "url": "http://example.com/test3",
    "normalized_url": "http example com test3",
    "title": "Title3",
    "normalized_title": "title3",
    "content": "Content3",
    "normalized_content": "content3",
    "vector": "'com':4A 'content3':6C 'example':3A 'http':2A 'test3':5A 'title3':1A",
}


SERIALIZED_CRAWLER_STATS = [
    {"doc_count": 23, "freq": "M", "indexing_speed": 2, "queued_url": 24, "t": now_str},
    {"doc_count": 33, "freq": "D", "indexing_speed": 4, "queued_url": 34, "t": now_str},
]


class RestAPITest:
    maxDiff = None

    def setUp(self):
        self.collection = Collection.create_default()
        self.collection2 = Collection.objects.create(name="Test Collection 2", unlimited_regex="http://example.com/.*")
        self.client = Client(HTTP_USER_AGENT="Mozilla/5.0")
        self.user = User.objects.create_user(username="admin", password="admin", is_superuser=True)
        self.user.save()

        self.client.login(username="admin", password="admin")

        self.doc1 = Document.objects.wo_content().create(
            url="http://127.0.0.1/test",
            normalized_url="http test",
            title="Title",
            normalized_title="title",
            content="Content",
            normalized_content="content",
            crawl_first=now,
            crawl_last=now,
            lang_iso_639_1="en",
            mimetype="text/html",
            collection=self.collection,
        )
        self.doc2 = Document.objects.wo_content().create(
            url="http://127.0.0.1/test2",
            normalized_url="http test2",
            title="Title2",
            normalized_title="title2",
            content="Other Content2",
            normalized_content="other content2",
            crawl_first=now,
            crawl_last=now,
            lang_iso_639_1="en",
            mimetype="image/png",
            collection=self.collection,
        )
        self.doc3 = Document.objects.wo_content().create(
            url="http://example.com/test3",
            normalized_url="http example com test3",
            title="Title3",
            normalized_title="title3",
            content="Content3",
            normalized_content="content3",
            crawl_first=now,
            crawl_last=now,
            lang_iso_639_1="en",
            mimetype="text/html",
            collection=self.collection2,
        )

        self.crawler_stat1 = CrawlerStats.objects.create(t=now, doc_count=23, queued_url=24, indexing_speed=2, freq="M")
        self.crawler_stat2 = CrawlerStats.objects.create(t=now, doc_count=33, queued_url=34, indexing_speed=4, freq="D")

        self.tag = Tag.objects.create(name="Group")
        self.subtag = Tag.objects.create(name="Sub Tag", parent=self.tag)
        self.doc2.tags.set([self.subtag])

        self.test_webhook = Webhook.objects.create(name="Test Webhook", url="http://test.com/webhook")

        self.mime_plugin_builtin = MimePlugin.objects.create(
            name="Built-in Handler",
            description="A built-in MIME handler",
            script="echo 'builtin'",
            mimetype_re="^text/plain$",
            builtin=True,
            enabled=True,
        )
        self.mime_plugin_custom = MimePlugin.objects.create(
            name="Custom Handler",
            description="A custom MIME handler",
            script="echo 'custom'",
            mimetype_re="^application/json$",
            builtin=False,
            enabled=True,
        )


class APIQueryTest(RestAPITest, TransactionTestCase):
    def test_document_list(self):
        response = self.client.get("/api/document/")
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 3)
        self.assertEqual(data["next"], None)
        self.assertEqual(data["previous"], None)

        doc_ids = [result["id"] for result in data["results"]]
        self.assertIn(self.doc1.id, doc_ids)
        self.assertIn(self.doc2.id, doc_ids)
        self.assertIn(self.doc3.id, doc_ids)

    def test_document_detail(self):
        response = self.client.get(f"/api/document/{self.doc1.id}/")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content), SERIALIZED_DOC1 | {"id": self.doc1.id, "collection": self.collection.id}
        )

    def test_stats_list(self):
        response = self.client.get("/api/stats/")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 2,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_CRAWLER_STATS[0] | {"id": self.crawler_stat1.id},
                    SERIALIZED_CRAWLER_STATS[1] | {"id": self.crawler_stat2.id},
                ],
            },
        )

    def test_stats_detail(self):
        response = self.client.get(f"/api/stats/{self.crawler_stat1.id}/")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            SERIALIZED_CRAWLER_STATS[0] | {"id": self.crawler_stat1.id},
        )

    def test_stats_list_filter(self):
        response = self.client.get("/api/stats/?freq=M")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_CRAWLER_STATS[0] | {"id": self.crawler_stat1.id},
                ],
            },
        )

    @mock.patch("os.statvfs")
    def test_hdd_stats(self, statvfs):
        fakevfs = namedtuple(
            "fakevfs",
            [
                "f_bsize",
                "f_frsize",
                "f_blocks",
                "f_bfree",
                "f_bavail",
                "f_files",
                "f_ffree",
                "f_favail",
                "f_flag",
                "f_namemax",
            ],
        )
        statvfs.side_effect = lambda x: fakevfs(
            f_bsize=4096,
            f_frsize=4096,
            f_blocks=51081736,
            f_bfree=30397904,
            f_bavail=27784887,
            f_files=13049856,
            f_ffree=11610989,
            f_favail=11610989,
            f_flag=4096,
            f_namemax=255,
        )
        response = self.client.get("/api/hdd_stats/")
        self.assertEqual(response.status_code, 200, response.content)

        content = json.loads(response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "db": content["db"],
                "free": 113806897152,
                "html": content["html"],
                "other": content["other"],
                "screenshots": content["screenshots"],
            },
        )

    def test_lang_stats(self):
        response = self.client.get("/api/lang_stats/")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(json.loads(response.content), [{"doc_count": 3, "lang": "En 🇬🇧"}])

    def test_mime_stats(self):
        response = self.client.get("/api/mime_stats/")
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            [{"mimetype": "🌐 text/html", "doc_count": 2}, {"mimetype": "🖼️ image/png", "doc_count": 1}],
            response.content,
        )

    def test_search(self):
        response = self.client.post("/api/search/", {"query": "content"})
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC1
                    | {
                        "collection": self.collection.id,
                        "id": self.doc1.id,
                        "score": 0.12158542,
                    }
                ],
            },
        )

    def test_search_has_params(self):
        response = self.client.post("/api/search/", {})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            json.loads(response.content),
            {"non_field_errors": ['At least "query" or "adv_params" field must be provided.']},
        )

    def test_search_title(self):
        response = self.client.post(
            "/api/search/",
            {"adv_params": [{"field": "content", "term": "Content2"}]},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC2
                    | {
                        "collection": self.collection.id,
                        "id": self.doc2.id,
                        "score": 1.0,
                    }
                ],
            },
        )

    def test_search_hidden(self):
        self.doc2.hidden = True
        self.doc2.save()

        response = self.client.post("/api/search/", {"query": "http"})
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 2,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC1
                    | {
                        "collection": self.collection.id,
                        "id": self.doc1.id,
                        "score": 0.6079271,
                    },
                    SERIALIZED_DOC3
                    | {
                        "collection": self.collection2.id,
                        "id": self.doc3.id,
                        "score": 0.6079271,
                    },
                ],
            },
        )

    def test_search_hidden_included(self):
        self.doc2.hidden = True
        self.doc2.save()

        response = self.client.post("/api/search/", {"query": "http", "include_hidden": True})
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 3,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC1
                    | {
                        "collection": self.collection.id,
                        "id": self.doc1.id,
                        "score": 0.6079271,
                    },
                    SERIALIZED_DOC2
                    | {
                        "collection": self.collection.id,
                        "id": self.doc2.id,
                        "score": 0.6079271,
                        "hidden": True,
                    },
                    SERIALIZED_DOC3
                    | {
                        "collection": self.collection2.id,
                        "id": self.doc3.id,
                        "score": 0.6079271,
                    },
                ],
            },
        )

    def test_search_tags(self):
        response = self.client.post(
            "/api/search/",
            {"adv_params": [{"field": "tag", "term": "Sub Tag", "operator": "equal"}]},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC2
                    | {
                        "collection": self.collection.id,
                        "id": self.doc2.id,
                        "score": 1.0,
                    },
                ],
            },
        )

    def test_search_subtags(self):
        response = self.client.post(
            "/api/search/",
            {"adv_params": [{"field": "tag", "term": "Group", "operator": "equal"}]},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.assertEqual(
            json.loads(response.content),
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    SERIALIZED_DOC2
                    | {
                        "collection": self.collection.id,
                        "id": self.doc2.id,
                        "score": 1.0,
                    },
                ],
            },
        )

    def test_document_tags_str_update(self):
        response = self.client.patch(
            f"/api/document/{self.doc1.id}/", {"tags": ["tag1", "tag2"]}, content_type="application/json"
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.doc1.refresh_from_db()
        tags = list(self.doc1.tags.values_list("name", flat=True))
        self.assertEqual(tags, ["tag1", "tag2"])

    def test_document_tags_pk_update(self):
        tag = Tag.objects.create(name="test tag")
        response = self.client.patch(
            f"/api/document/{self.doc1.id}/", {"tags": [tag.pk]}, content_type="application/json"
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.doc1.refresh_from_db()
        tags = list(self.doc1.tags.values_list("name", flat=True))
        self.assertEqual(tags, ["test tag"])

    def test_document_normalize(self):
        response = self.client.patch(
            f"/api/document/{self.doc1.id}/", {"title": "tïtle", "content": "contènt"}, content_type="application/json"
        )
        self.assertEqual(response.status_code, 200, response.content)
        self.doc1.refresh_from_db()
        self.assertEqual(self.doc1.normalized_title, "title")
        self.assertEqual(self.doc1.normalized_content, "content")

    def test_document_create_prohibited(self):
        response = self.client.post("/api/document/", {"url": "http://127.0.0.1/"}, content_type="application/json")
        self.assertEqual(response.status_code, 405, response.content)

    def test_mime_plugin_list(self):
        response = self.client.get("/api/mime_plugin/")
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 2)
        self.assertTrue(any(handler["name"] == "Built-in Handler" for handler in data["results"]))
        self.assertTrue(any(handler["name"] == "Custom Handler" for handler in data["results"]))

    def test_mime_plugin_detail(self):
        response = self.client.get(f"/api/mime_plugin/{self.mime_plugin_custom.id}/")
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["name"], "Custom Handler")
        self.assertEqual(data["builtin"], False)

    def test_mime_plugin_create(self):
        response = self.client.post(
            "/api/mime_plugin/",
            {
                "name": "New Handler",
                "description": "A new handler",
                "script": "echo 'new'",
                "mimetype_re": "^image/jpeg$",
                "enabled": True,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["name"], "New Handler")
        self.assertEqual(data["builtin"], False)

    def test_mime_plugin_update_custom(self):
        response = self.client.patch(
            f"/api/mime_plugin/{self.mime_plugin_custom.id}/",
            {"description": "Updated description"},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["description"], "Updated description")

    def test_mime_plugin_update_builtin_forbidden(self):
        response = self.client.patch(
            f"/api/mime_plugin/{self.mime_plugin_builtin.id}/",
            {"description": "Should not work"},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 403, response.content)
        data = json.loads(response.content)
        self.assertIn("for built-in MIME handlers", data["detail"])

    def test_mime_plugin_delete_custom(self):
        response = self.client.delete(f"/api/mime_plugin/{self.mime_plugin_custom.id}/")
        self.assertEqual(response.status_code, 204, response.content)

    def test_mime_plugin_delete_builtin_forbidden(self):
        response = self.client.delete(f"/api/mime_plugin/{self.mime_plugin_builtin.id}/")
        self.assertEqual(response.status_code, 403, response.content)
        data = json.loads(response.content)
        self.assertIn("Cannot delete built-in MIME handlers", data["detail"])

    def test_mime_plugin_builtin_readonly(self):
        response = self.client.post(
            "/api/mime_plugin/",
            {
                "name": "Test Handler",
                "script": "echo 'test'",
                "mimetype_re": "^text/test$",
                "builtin": True,
                "enabled": True,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["builtin"], False)

    def test_search_collection_filter(self):
        response = self.client.post("/api/search/", {"query": "content", "collection": self.collection.id})
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], self.doc1.id)

    def test_search_collection_filter_different(self):
        response = self.client.post("/api/search/", {"query": "content3", "collection": self.collection2.id})
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], self.doc3.id)

    def test_search_no_collection_filter(self):
        response = self.client.post("/api/search/", {"query": "http"})
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 3)
        doc_ids = [result["id"] for result in data["results"]]
        self.assertIn(self.doc1.id, doc_ids)
        self.assertIn(self.doc2.id, doc_ids)
        self.assertIn(self.doc3.id, doc_ids)

    def test_search_collection_nonexistent(self):
        response = self.client.post("/api/search/", {"query": "content", "collection": 99999})
        self.assertEqual(response.status_code, 400, response.content)
        data = json.loads(response.content)
        self.assertIn("Collection with id 99999 does not exist", data["collection"][0])

    def test_collection_list(self):
        response = self.client.get("/api/collection/")
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["count"], 2)

        collection_names = [result["name"] for result in data["results"]]
        self.assertEqual(len(collection_names), 2)
        self.assertIn(self.collection.name, collection_names)
        self.assertIn("Test Collection 2", collection_names)

    def test_collection_detail(self):
        response = self.client.get(f"/api/collection/{self.collection.id}/")
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["id"], self.collection.id)
        self.assertEqual(data["name"], self.collection.name)

    def test_collection_create(self):
        response = self.client.post(
            "/api/collection/",
            {"name": "New Collection", "unlimited_regex": "http://test.com/.*"},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["name"], "New Collection")
        self.assertEqual(data["unlimited_regex"], "http://test.com/.*")
        self.assertEqual(data["unlimited_regex_pg"], "http://test.com/.*")

    def test_collection_update(self):
        response = self.client.patch(
            f"/api/collection/{self.collection2.id}/",
            {"name": "Updated Collection", "unlimited_regex": "http://updated.com/.*\nhttp://other.com/.*"},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["name"], "Updated Collection")
        self.assertEqual(data["unlimited_regex"], "http://updated.com/.*\nhttp://other.com/.*")
        self.assertEqual(data["unlimited_regex_pg"], "(http://updated.com/.*|http://other.com/.*)")

    def test_collection_delete(self):
        new_collection = Collection.objects.create(name="To Delete", unlimited_regex="http://delete.com/.*")
        response = self.client.delete(f"/api/collection/{new_collection.id}/")
        self.assertEqual(response.status_code, 204, response.content)
        self.assertFalse(Collection.objects.filter(id=new_collection.id).exists())

    def test_collection_create_with_tags_webhooks(self):
        response = self.client.post(
            "/api/collection/",
            {
                "name": "Collection with Relations",
                "unlimited_regex": "http://test.com/.*",
                "tags": [self.tag.id, self.subtag.id],
                "webhooks": [self.test_webhook.id],
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["name"], "Collection with Relations")
        self.assertEqual(sorted(data["tags"]), sorted([self.tag.id, self.subtag.id]))
        self.assertEqual(data["webhooks"], [self.test_webhook.id])

    def test_collection_update_tags_webhooks(self):
        response = self.client.patch(
            f"/api/collection/{self.collection2.id}/",
            {"tags": [self.subtag.id], "webhooks": [self.test_webhook.id]},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["tags"], [self.subtag.id])
        self.assertEqual(data["webhooks"], [self.test_webhook.id])

    def test_queue_urls_basic(self):
        initial_count = Document.objects.count()
        response = self.client.post(
            "/api/queue/",
            {
                "urls": ["http://newsite.com/page1", "http://newsite.com/page2"],
                "collection": self.collection.id,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["message"], "2 URLs queued successfully")
        self.assertEqual(len(data["queued_urls"]), 2)
        self.assertIn("http://newsite.com/page1", data["queued_urls"])
        self.assertIn("http://newsite.com/page2", data["queued_urls"])
        self.assertEqual(data["collection"], self.collection.id)
        self.assertEqual(Document.objects.count(), initial_count + 2)

    def test_queue_urls_unlimited_scope(self):
        response = self.client.post(
            "/api/queue/",
            {
                "urls": ["http://newsite.com/page"],
                "collection": self.collection.id,
                "crawl_scope": "unlimited",
                "show_on_homepage": False,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)
        data = json.loads(response.content)
        self.assertEqual(data["crawl_scope"], "unlimited")
        self.assertEqual(data["show_on_homepage"], False)

        self.collection.refresh_from_db()
        self.assertIn("^https?://newsite\\.com/.*", self.collection.unlimited_regex)

    def test_queue_urls_limited_scope(self):
        response = self.client.post(
            "/api/queue/",
            {
                "urls": ["http://testsite.org/"],
                "collection": self.collection.id,
                "crawl_scope": "limited",
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 201, response.content)

        self.collection.refresh_from_db()
        self.assertIn("^https?://testsite\\.org/.*", self.collection.limited_regex)

    def test_queue_urls_invalid_url(self):
        response = self.client.post(
            "/api/queue/",
            {
                "urls": ["not-a-valid-url", "http://valid.com"],
                "collection": self.collection.id,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400, response.content)
        data = json.loads(response.content)
        self.assertIn("urls", data)

    def test_queue_urls_nonexistent_collection(self):
        response = self.client.post(
            "/api/queue/",
            {
                "urls": ["http://test.com"],
                "collection": 99999,
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400, response.content)
        data = json.loads(response.content)
        self.assertIn("collection", data)
